# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from collections import Counter

from traits.api import (
    Any, Bool, DelegatesTo, Enum, Event, HasStrictTraits, Instance, Int,
    Property, String, WeakRef, on_trait_change)

from karabo.common import const
from karabo.native import Hash
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.singletons.api import get_network, get_topology

from .binding_types import (
    BaseBinding, BindingRoot, FloatBinding, PipelineOutputBinding, SlotBinding,
    VectorBinding)
from .enums import ONLINE_STATUSES, SCHEMA_STATUSES, ProxyStatus

_NOCAST_BINDINGS = (FloatBinding, VectorBinding)


class BaseDeviceProxy(HasStrictTraits):
    """Base class for proxying device classes, device instances, or project
    devices.
    """
    # The root of the data binding
    binding = Instance(BindingRoot)
    # The root's `state` property, if existent
    state_binding = Instance(BaseBinding)
    # The root's `lockedBy` property, if existent
    locked_binding = Instance(BaseBinding)
    # ID of the server hosting this class
    server_id = String
    # True when the device is online
    online = Property(Bool, depends_on='status')
    # The current status
    status = Enum(*ProxyStatus)
    # An event which fires when the schema is updated or otherwise changes
    schema_update = DelegatesTo('binding')

    @on_trait_change('binding.schema_update')
    def _binding_changed(self):
        """When `binding` is assigned to or its schema changes, refresh
        `state_binding`
        """
        self.state_binding = self.get_property_binding('state')
        self.locked_binding = self.get_property_binding('lockedBy')

    def _get_online(self):
        return self.status in ONLINE_STATUSES

    def _status_default(self):
        return ProxyStatus.OFFLINE

    def get_property_binding(self, path):
        """Return the ``BaseBinding``-derived instance for the proxy property
        denoted by `path`.

        NOTE: Returns `None` if the binding is not set, or the property does
        not yet exist in the binding (eg: when awaiting schema injection)
        """
        prop = self.binding
        for part in path.split('.'):
            if prop is None:
                break
            prop = getattr(prop.value, part, None)
        return prop

    def __getitem__(self, path):
        return self.get_property_binding(path)


class DeviceProxy(BaseDeviceProxy):
    """The proxy for device instances and project devices. Instances may come
    on and offline.
    """
    # ID of the device we are proxying
    device_id = String
    # An event which fires when the configuration is updated from the network
    config_update = Event
    # The system tree node for this device
    topology_node = WeakRef('karabogui.topology.system_tree.SystemTreeNode',
                            allow_none=True)

    # A reference counter for tracking the monitoring
    _monitor_count = Int(0)
    # A counter for pipeline connection objects
    _pipeline_subscriptions = Instance(Counter, args=())

    # -----------------------------------------------------------------------
    # Traits methods

    def _config_update_fired(self):
        if self.status is ProxyStatus.SCHEMA:
            self.status = ProxyStatus.ALIVE
        if self.status is ProxyStatus.ALIVE and self._monitor_count > 0:
            self.status = ProxyStatus.MONITORING

    def _schema_update_fired(self):
        if self.status is ProxyStatus.ONLINEREQUESTED:
            if self._monitor_count > 0:
                self._start_monitoring()
                # reestablish the pipelining
                for path in self._pipeline_subscriptions:
                    get_network().onSubscribeToOutput(self.device_id, path,
                                                      True)
            self.status = ProxyStatus.SCHEMA
        elif self.status in (ProxyStatus.ALIVE, ProxyStatus.MONITORING):
            get_network().onGetDeviceConfiguration(self.device_id)

    def _status_changed(self, new):
        if new is ProxyStatus.ONLINE and self._monitor_count > 0:
            self.refresh_schema()

    def __monitor_count_changed(self, old, new):
        if self.topology_node is not None:
            self.topology_node.monitoring = (new > 0)

        if old == 0 and new == 1:
            broadcast_event(KaraboEvent.StartMonitoringDevice,
                            {'device_id': self.device_id})
        elif old == 1 and new == 0:
            broadcast_event(KaraboEvent.StopMonitoringDevice,
                            {'device_id': self.device_id})

    def _topology_node_changed(self, new):
        if new is not None:
            new.monitoring = (self._monitor_count > 0)

    # -----------------------------------------------------------------------
    # Public interface

    def add_monitor(self):
        """Ask the GUI server to begin monitoring this device
        """
        ignored_statuses = (ProxyStatus.OFFLINE,
                            ProxyStatus.ONLINEREQUESTED)

        self._monitor_count += 1
        if self._monitor_count == 1:
            if self.status is ProxyStatus.ONLINE:
                self.refresh_schema()
            elif self.status not in ignored_statuses:
                self._start_monitoring()

    def remove_monitor(self):
        """Ask the GUI server to stop monitoring this device
        """
        assert self._monitor_count > 0
        self._monitor_count -= 1
        if self._monitor_count == 0:
            self._stop_monitoring()

    def connect_pipeline(self, path):
        """Ask the GUI server to subscribe to a pipeline output
        """
        if path not in self._pipeline_subscriptions:
            get_network().onSubscribeToOutput(self.device_id, path, True)

        # NOTE: We fill a counter object with the path, as we might be
        # interested in multiple values from an output channel
        self._pipeline_subscriptions[path] += 1

    def disconnect_pipeline(self, path):
        """Ask the GUI server to unsubscribe from a pipeline output
        """
        assert path in self._pipeline_subscriptions

        self._pipeline_subscriptions[path] -= 1
        # NOTE: Only if we fully removed all interested properties
        # we unsubscribe from the output channel
        if self._pipeline_subscriptions[path] == 0:
            del self._pipeline_subscriptions[path]
            get_network().onSubscribeToOutput(self.device_id, path, False)

    def publish_historic_data(self, path, data):
        """A callback which routes historic data received from the GUI server
        to the appropriate binding object.
        """
        binding = self.get_property_binding(path)
        if binding is not None:
            # NOTE: `historic_data` is an Event trait which does not store
            # any values assigned to it! The value is only available to
            # traits notification handlers which are listening to the event.
            binding.historic_data = data

    def refresh_schema(self):
        """Request a recent schema for this device instance
        """
        if self.status is not ProxyStatus.ONLINEREQUESTED:
            get_network().onGetDeviceSchema(self.device_id)
            self.status = ProxyStatus.ONLINEREQUESTED

    # -----------------------------------------------------------------------
    # Private interface

    def _start_monitoring(self):
        get_network().onStartMonitoringDevice(self.device_id)

    def _stop_monitoring(self):
        """Stop monitoring of the device proxy"""
        get_network().onStopMonitoringDevice(self.device_id)
        if self.status in (ProxyStatus.ALIVE, ProxyStatus.MONITORING):
            self.status = ProxyStatus.ONLINE

    def __repr__(self):
        return f"<{self.__class__.__name__} deviceId={self.device_id}>"


class DeviceClassProxy(BaseDeviceProxy):
    """The proxy for device classes. Used for editing before instantiation.
    """
    # An event which fires when the configuration has been applied
    config_update = Event

    def _status_default(self):
        topology = get_topology()
        server_key = f'server.{self.server_id}'
        attributes = topology.get_attributes(server_key)
        if attributes is None:
            return ProxyStatus.NOSERVER
        elif self.binding.class_id not in attributes.get('deviceClasses', []):
            return ProxyStatus.NOPLUGIN
        return ProxyStatus.OFFLINE

    def _schema_update_fired(self):
        if self.status is ProxyStatus.REQUESTED:
            self.status = self._status_default()

    def refresh_schema(self):
        """Request a recent schema for this device class
        """
        if self.status is not ProxyStatus.REQUESTED:
            get_network().onGetClassSchema(self.server_id,
                                           self.binding.class_id)
            self.status = ProxyStatus.REQUESTED

    def __repr__(self):
        return (f"<{self.__class__.__name__} "
                f"classId={getattr(self.binding, 'class_id', None)} "
                f"serverId={self.server_id}>")


class ProjectDeviceProxy(DeviceClassProxy):
    """A device class proxy used in ProjectDeviceInstance
    """
    device_id = String

    def update_status(self):
        """Request a recalculating of the device status
        """
        self.status = self._status_default()

    def ensure_class_schema(self):
        """Ensure the class schema of the project device proxy"""
        topology = get_topology()
        topology.ensure_proxy_class_schema(
            self.device_id, self.server_id, self.binding.class_id)

    def __repr__(self):
        return f"<{self.__class__.__name__} deviceId={self.device_id}>"


class PropertyProxy(HasStrictTraits):
    """A proxy for a single device property

    :param path: Full path of the property
    :param key: Full 'key' of the property: <device ID>.<path>
    :param binding: The binding for the property
    :param value: The value for the property (from the binding instance)

    :param edit_value: A user-entered value
    :param root_proxy: Parent device proxy
    :param pipeline_parent_path: Potential parent path if `binding` is a
                                 child of a Pipeline Output

    """
    path = String
    key = Property(String)
    binding = Instance(BaseBinding)
    value = Property(depends_on='binding.value')
    edit_value = Any
    root_proxy = Instance(BaseDeviceProxy, allow_none=False)
    pipeline_parent_path = String

    # More private members
    # ---------------------

    visible = Bool(False)
    # Check whether we are existing or not
    existing = Bool(True)

    # An extra binding instance for `edit_value` validation
    _edit_binding = Instance(BaseBinding)

    # -----------------------------------------------------------------------
    # Traits methods

    def _get_key(self):
        if not isinstance(self.root_proxy, DeviceProxy):
            return self.path
        return self.root_proxy.device_id + '.' + self.path

    def _get_value(self):
        binding = self.binding
        if binding is not None:
            return binding.value

    def _set_value(self, value):
        binding = self.binding
        if binding is not None:
            binding.value = value

    def _pipeline_parent_path_default(self):
        if self.binding is None:
            return ''

        def _gen_parents(p):
            if '.' in p:
                p = p.rsplit('.', 1)[0]
                yield p
                yield from _gen_parents(p)

        # Search the parents of this node (via path manipulation) for pipeline
        # outputs. Return the path of that parent if found.
        for path in _gen_parents(self.path):
            binding = self.root_proxy.get_property_binding(path)
            if isinstance(binding, PipelineOutputBinding):
                return path
        return ''  # No parent pipeline output node found

    def _pipeline_parent_path_changed(self, new):
        """Make sure to connect a pipeline if a schema arrives after we start
        monitoring!
        """
        if new == '':
            return

        if self.visible and isinstance(self.root_proxy, DeviceProxy):
            # Call blindling, the device proxy will keep things orderly
            self.root_proxy.connect_pipeline(new)

    @on_trait_change('root_proxy.schema_update,path')
    def _binding_update(self):
        # Clear some edit related traits no matter what
        self._edit_binding = None
        if self.root_proxy is None:
            return

        self.binding = self.root_proxy.get_property_binding(self.path)
        if self.root_proxy.status in SCHEMA_STATUSES:
            self.existing = self.binding is not None
            self.pipeline_parent_path = self._pipeline_parent_path_default()

    def _edit_value_changed(self, value):
        """When `edit_value` changes, set it to a binding object to validate
        its value.
        """
        if value is None:
            return

        if self._edit_binding is None:
            klass = type(self.binding)
            self._edit_binding = klass()

        # Validate (and mutate in some cases, like arrays)
        self._edit_binding.value = value
        # Replace quietly but don't send no cast bindings to gui
        self.trait_setq(edit_value=self._edit_binding.value)
        if not isinstance(self.binding, _NOCAST_BINDINGS):
            self.binding.config_update = True

    # -----------------------------------------------------------------------
    # Public methods

    def execute(self):
        """Call a slot on a remote device.
        """
        if not (isinstance(self.root_proxy, DeviceProxy) and
                isinstance(self.binding, SlotBinding)):
            return

        state = self.root_proxy.state_binding.value
        if self.binding.is_allowed(state):
            # Macro Slots are not guaranteed to reply immediately
            # we do not expect a reply from them.
            # XXX: this protection should be implemented on macro side
            t_node = self.root_proxy.topology_node
            ignore_timeouts = False
            if t_node:
                # ignore timeouts if the root_proxy is a macro.
                ignore_timeouts = t_node.attributes.get('type') == 'macro'
            network = get_network()
            network.onExecute(self.root_proxy.device_id,
                              self.path, ignore_timeouts)

            info = Hash("type", "execute",
                        "instanceId", self.root_proxy.device_id,
                        "slot", self.path)
            network.onInfo(info)

    def get_device_value(self):
        """Return the value stored in the device configuration, rather
        than that which is stored in `binding`.

        For a device class, return the defaul value, if one exists.
        Returns None if there is no value.
        """
        if self.binding is None:
            return None

        if isinstance(self.root_proxy, DeviceProxy):
            return self.binding.value
        else:
            attrs = self.binding.attributes
            return attrs.get(const.KARABO_SCHEMA_DEFAULT_VALUE)

    def get_history(self, time_start, time_end, max_value_count=-1):
        """Request the historical values of `binding` between `time_start` and
        `time_end` (XXX: inclusive?).

        NOTE: The reply comes via `self.binding`s `historic_data` Event trait.
        """
        if (self.binding is None or
                not isinstance(self.root_proxy, DeviceProxy)):
            return

        get_network().onGetPropertyHistory(self.root_proxy.device_id,
                                           self.path, time_start, time_end,
                                           max_value_count)

    def revert_edit(self):
        """Revert any local edits made to a property.
        """
        self.edit_value = None
        if self.binding is not None:
            self.binding.config_update = True

    def start_monitoring(self):
        """Tell the GUI server to start monitoring the device for our property
        """
        self.visible = True

        if not isinstance(self.root_proxy, DeviceProxy):
            return

        self.root_proxy.add_monitor()
        if self.pipeline_parent_path:
            self.root_proxy.connect_pipeline(self.pipeline_parent_path)

    def stop_monitoring(self):
        """Tell the GUI server to stop monitoring the device for our property
        """
        self.visible = False

        if not isinstance(self.root_proxy, DeviceProxy):
            return

        self.root_proxy.remove_monitor()
        if self.pipeline_parent_path:
            self.root_proxy.disconnect_pipeline(self.pipeline_parent_path)

    # -----------------------------------------------------------------------
    # Private and internal methods

    def __eq__(self, other):
        """Enable comparison of instances"""
        if not isinstance(other, PropertyProxy):
            return False
        return (self.root_proxy is other.root_proxy and
                self.path == other.path)

    def __hash__(self):
        """Traits objects need to be hashable, and defining __eq__ clears the
        default __hash__ implementation.
        """
        return id(self)

    def __repr__(self):
        return f"<{self.__class__.__name__} key={self.key}>"



from collections import namedtuple
from pathlib import Path

from qtpy.QtCore import QObject, Signal

from karabo.common.scenemodel.api import SceneTargetWindow
from karabogui.events import KaraboEvent
from karabogui.programs.concert import create_concert, run_concert
from karabogui.testing import singletons, system_hash
from karabogui.topology.api import SystemTopology

namespace = namedtuple("namespace", "yaml_file")

DATA = {
    "domain": "CAS_INTERNAL",
    "scenes": [{"uuid": "9d26d07b-d79e-4f71-99f1-469344da89ed",
                "x": 1458, "y": 93},
               {"uuid": "9b73efec-4523-4a47-b88f-90edc2e9ed89",
                "x": 2332, "y": 361}],
    "host": "exflqr46957", "port": 44444}


def test_run_concert(gui_app, mocker):
    file_path = Path(__file__).parent / "concert.yml"
    args = namespace(file_path)

    mock_concert = mocker.patch("karabogui.programs.concert.create_concert",
                                return_value=(False, "dummyApp"))
    run_concert(args)
    assert mock_concert.call_count == 1
    mock_concert.assert_called_with(DATA)


class NetworkMock(QObject):
    """A basic mock to avoid using the singleton class"""
    signalServerConnectionChanged = Signal(bool)
    signalReceivedData = Signal(object)
    signalNetworkPerformance = Signal(float, bool)

    def __init__(self, parent=None):
        super().__init__(parent)


def test_concert(gui_app, mocker):
    network = NetworkMock()
    network.connectToServerDirectly = mocker.Mock()
    network.connectToServerDirectly.return_value = True
    network.onSubscribeLogs = mocker.Mock()
    topology = SystemTopology()

    path = "karabogui.programs.concert.broadcast_event"
    concert_app = mocker.patch("karabogui.programs.concert.create_gui_app")
    concert_app.return_value = gui_app
    assert gui_app.organizationName() == "TestFacility"
    # Patch the gui_app to avoid collisions between tests
    with singletons(network=network, topology=topology):
        broadcast = mocker.patch(path)
        success, app = create_concert(DATA)
        concert_app.assert_called_once()
        assert success

        network.signalServerConnectionChanged.emit(True)
        network.onSubscribeLogs.assert_not_called()

        # Initialize the trees
        topology.initialize(system_hash())
        expected_args1 = {
            "name": "CAS_INTERNAL-9d26d07b-d79e-4f71-99f1-469344da89ed",
            "target": "9d26d07b-d79e-4f71-99f1-469344da89ed",
            "position": [1458, 93],
            "target_window": SceneTargetWindow.Dialog}
        expected_args2 = {
            "name": "CAS_INTERNAL-9b73efec-4523-4a47-b88f-90edc2e9ed89",
            "target": "9b73efec-4523-4a47-b88f-90edc2e9ed89",
            "position": [2332, 361],
            "target_window": SceneTargetWindow.Dialog}

        assert broadcast.call_count == 2
        args1, args2 = broadcast.call_args_list
        assert args1[0] == (KaraboEvent.OpenSceneLink, expected_args1)
        assert args2[0] == (KaraboEvent.OpenSceneLink, expected_args2)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import os

from qtpy import uic
from qtpy.QtCore import Qt, Slot
from qtpy.QtGui import QColor, QIcon, QPixmap
from qtpy.QtWidgets import QColorDialog, QDialog


class GraphViewDialog(QDialog):

    def __init__(self, config={}, parent=None):
        super().__init__(parent)
        ui_path = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                               "view.ui")
        uic.loadUi(ui_path, self)

        self.graph_title = config.get("title", "")
        self.ui_title.setText(self.graph_title)
        self.graph_bg_color = config.get("background", "transparent")
        self.set_text_background_button()
        self.ui_cb_background.setChecked(
            self.graph_bg_color != "transparent")

    def set_text_background_button(self):
        pixmap = QPixmap(24, 16)
        pixmap.fill(QColor(self.graph_bg_color))
        self.ui_pb_background.setIcon(QIcon(pixmap))

    @Slot(int)
    def on_ui_cb_background_stateChanged(self, state):
        if state != Qt.Checked:
            self.graph_bg_color = "transparent"

    @Slot(bool)
    def on_ui_cb_background_toggled(self, checked):
        self.ui_pb_background.setEnabled(checked)
        if not checked:
            self.graph_bg_color = "transparent"

    @Slot()
    def on_ui_pb_background_clicked(self):
        color = QColorDialog.getColor(QColor(self.graph_bg_color))
        if color.isValid():
            self.graph_bg_color = color.name()
            self.set_text_background_button()

    @property
    def settings(self):
        config = {
            "background": self.graph_bg_color,
            "title": self.ui_title.text(),
        }

        return config

    @staticmethod
    def get(configuration, parent=None):
        dialog = GraphViewDialog(configuration, parent)
        result = dialog.exec() == QDialog.Accepted
        content = {}
        content.update(dialog.settings)

        return content, result



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from traits.api import Instance, Int

from karabo.common.scenemodel.api import BaseWidgetObjectData
from karabo.native import AccessMode, Configurable, String
from karabogui.binding.api import (
    DeviceClassProxy, PropertyProxy, ProxyStatus, StringBinding, build_binding)
from karabogui.testing import flushed_registry, set_proxy_value

from ..baselineedit import BaseLineEditController
from ..registry import register_binding_controller


class Object(Configurable):
    string = String(accessMode=AccessMode.RECONFIGURABLE)


class UniqueWidgetModel(BaseWidgetObjectData):
    pass


def _define_binding_classes():
    @register_binding_controller(klassname="NewEdit",
                                 binding_type=StringBinding,
                                 can_edit=True, priority=-30)
    class CustomLineEdit(BaseLineEditController):
        model = Instance(UniqueWidgetModel, args=())

        counter = Int(0)

        def onText(self, text):
            super().onText(text)
            self.counter += 1

    return {
        "CustomLineEdit": CustomLineEdit,
    }


def test_custom_line_edit(gui_app):
    """Test the creation of a simple `BaseLineEditController`"""
    with flushed_registry():
        klasses = _define_binding_classes()

    # Save the classes so they only need to be declared once
    edit_controller = klasses["CustomLineEdit"]

    schema = Object.getClassSchema()
    binding = build_binding(schema)
    device = DeviceClassProxy(binding=binding, server_id="NoDeviceServer",
                              status=ProxyStatus.OFFLINE)
    proxy = PropertyProxy(root_proxy=device, path="string")
    assert proxy.binding is not None

    controller = edit_controller(proxy=proxy)
    controller.create(None)
    controller.set_read_only(False)
    assert controller.widget is not None
    assert controller.counter == 0

    controller.finish_initialization()
    assert controller.internal_widget.validator() is None
    set_proxy_value(proxy, "string", "Update")
    assert controller.internal_widget.text() == "Update"
    assert controller.counter == 0

    controller.internal_widget.setText("NoBeam")
    assert controller.counter == 1
    assert controller.internal_widget.text() == "NoBeam"

    assert controller.internal_widget.isEnabled()
    controller.setEnabled(False)
    assert not controller.internal_widget.isEnabled()
    controller.setEnabled(True)
    assert controller.internal_widget.isEnabled()



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on January 7, 2019
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from abc import abstractmethod

from traits.api import ABCHasStrictTraits

from karabogui.dialogs.api import LogDialog
from karabogui.itemtypes import NavigationItemTypes
from karabogui.logger import get_logger
from karabogui.request import retrieve_default_scene
from karabogui.util import move_to_cursor, version_compatible


class NavigationHandler(ABCHasStrictTraits):
    @abstractmethod
    def can_handle(self, info):
        """Check whether the click event can be handled."""

    @abstractmethod
    def handle(self, info, parent=None):
        """Handle the click event."""


class DeviceSceneHandler(NavigationHandler):
    """Device DoubleClick for the navigation view
    """

    def can_handle(self, info):
        navigation_type = info.get('type')
        if navigation_type is NavigationItemTypes.DEVICE:
            return True
        return False

    def handle(self, info, parent=None):
        device_id = info.get("deviceId")
        retrieve_default_scene(device_id)


class ServerLogHandler(NavigationHandler):
    """Device DoubleClick for the navigation view
    """

    def can_handle(self, info):
        navigation_type = info.get("type")
        if navigation_type is NavigationItemTypes.SERVER:
            return True
        return False

    def handle(self, info, parent=None):
        server_id = info.get("serverId")
        karabo_version = info.get("karaboVersion")

        if not version_compatible(karabo_version, 2, 14):
            get_logger().info(
                f"The server <b>{server_id}</b> cannot provide logs "
                f"since its framework version ({karabo_version}) is "
                "lower than 2.14.")
            return

        widget = LogDialog(server_id, parent=parent)
        move_to_cursor(widget)
        widget.show()
        widget.raise_()
        widget.activateWindow()



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on November 9, 2017
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import weakref
from pathlib import Path

from qtpy import uic
from qtpy.QtCore import QItemSelection, Qt, Slot
from qtpy.QtWidgets import (
    QDialog, QDialogButtonBox, QHBoxLayout, QPushButton, QVBoxLayout, QWidget)

import karabogui.icons as icons
from karabogui.controllers.edit.table import EditableTableElement

MAX_WIDTH = 1024
COLUMN_FACTOR = 120


class TableToolBar(QWidget):
    def __init__(self, controller, parent=None):
        super().__init__(parent=parent)
        self.setAttribute(Qt.WA_DeleteOnClose)
        self._controller = weakref.ref(controller)
        selection_model = controller.tableWidget().selectionModel()
        selection_model.selectionChanged.connect(self._table_selection_changed)

        self._move_up_button = QPushButton(icons.arrowFancyUp, "", self)
        self._move_down_button = QPushButton(icons.arrowFancyDown, "", self)
        self._add_button = QPushButton(icons.add, "", self)
        self._du_button = QPushButton(icons.editCopy, "", self)
        self._remove_button = QPushButton(icons.delete, "", self)

        self._move_up_button.clicked.connect(self._move_row_up)
        self._move_down_button.clicked.connect(self._move_row_down)
        self._add_button.clicked.connect(self._add_row_below)
        self._du_button.clicked.connect(self._duplicate_row)
        self._remove_button.clicked.connect(self._remove_row)

        layout = QVBoxLayout()
        layout.addWidget(self._move_up_button)
        layout.addWidget(self._move_down_button)
        layout.addWidget(self._add_button)
        layout.addWidget(self._du_button)
        layout.addWidget(self._remove_button)
        layout.addStretch(10)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        self._set_button_states()

    @property
    def controller(self):
        """Return the weakly stored controller variable"""
        return self._controller()

    @Slot()
    def _add_row_below(self):
        self.controller.add_row_below()
        self._set_button_states()

    @Slot()
    def _duplicate_row(self):
        self.controller.duplicate_row()
        self._set_button_states()

    @Slot()
    def _move_row_up(self):
        self.controller.move_row_up()
        self._set_button_states()

    @Slot()
    def _move_row_down(self):
        self.controller.move_row_down()

    @Slot()
    def _remove_row(self):
        self.controller.remove_row()
        self._set_button_states()

    @Slot(QItemSelection, QItemSelection)
    def _table_selection_changed(self, selected, deselected):
        self._set_button_states()

    def _set_button_states(self):
        model = self.controller.sourceModel()
        if model is None:
            # Note: This happens when the dialog is closed.
            return

        selection_model = self.controller.tableWidget().selectionModel()
        has_selection = selection_model.hasSelection()
        current_row = self.controller.currentIndex().row()
        self._move_up_button.setEnabled(has_selection and current_row > 0)
        self._move_down_button.setEnabled(has_selection and current_row <
                                          model.rowCount() - 1)
        self._du_button.setEnabled(has_selection)
        self._remove_button.setEnabled(has_selection)


class TableDialog(QDialog):
    def __init__(self, proxy, editable, parent=None):
        """A dialog to configure a VectorHash

        :param proxy: The `PropertyProxy` object representing the VectorHash
        :param editable: If the controller inside the dialog is editable
        :param parent: The parent of the dialog
        """
        super().__init__(parent)
        filepath = Path(__file__).parent / "table_view.ui"
        uic.loadUi(filepath, self)
        # XXX: Do not change the name of this variable since the
        # `EditDelegate` method `setModelData` refers to this dialog as the
        # passed `editor` and fetches the data of the `controller`
        self.controller = EditableTableElement(proxy=proxy)
        # This is parented by the `EditDelegate`, which means it will stay
        # alive very long!
        self.controller.create(parent)
        self.controller.set_read_only(not editable)
        self.controller.binding_update(proxy)
        self.controller.value_update(proxy)
        self.toolbar = None

        layout = QHBoxLayout()
        layout.setContentsMargins(2, 2, 2, 2)
        layout.addWidget(self.controller.widget)
        if editable:
            self.toolbar = TableToolBar(controller=self.controller)
            layout.addWidget(self.toolbar)
        # `empty_widget` is defined in table_edit.ui
        self.empty_widget.setLayout(layout)
        cancel_button = self.buttonBox.button(QDialogButtonBox.Cancel)
        cancel_button.setVisible(editable)

        width = 450
        height = 350

        binding = proxy.binding
        if binding is not None:
            columns = len(binding.bindings)
            # minimum width of 450, but can go up to `MAX_WIDTH`
            width = min(max(columns * COLUMN_FACTOR, width), MAX_WIDTH)
        self.resize(width, height)

    def done(self, result):
        self.controller.destroy_widget()
        self.toolbar = None
        return super().done(result)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from karabo.native import Hash
from karabogui.dialogs.api import DataViewDialog


def test_data_view(gui_app):
    data = Hash("Data", "Karabo")

    title = "This is the title"
    info = "This is the description info"
    dialog = DataViewDialog(title, info, data)
    assert dialog.windowTitle() == title
    assert dialog.ui_info.text() == info
    assert dialog.ui_text_info.toPlainText() == "\nData\nKarabo\n"



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest

from karabogui.programs.protocol_handler import (
    get_cinema_args, get_theatre_args, parse_url)


def test_parse_url():
    # Check URL with a domain and two scene UUIDs
    url = "karabo://cinema?domain=SQS&scene_uuid=xxx-xxx&scene_uuid=yyy-yyy"
    host, queries = parse_url(url)
    assert host == "cinema"
    assert queries == {"domain": ["SQS"],
                       "scene_uuid": ["xxx-xxx", "yyy-yyy"]}

    # Check URL with a domain, a scene UUID and a nosplash flag
    url = "karabo://cinema?domain=SQS&scene_uuid=xxx-xxx&nosplash"
    host, queries = parse_url(url)
    assert host == "cinema"
    assert queries == {"domain": ["SQS"],
                       "scene_uuid": ["xxx-xxx"],
                       "nosplash": ['']}

    # Check URL with a domain, a scene UUID, annd username/host/port
    url = "karabo://cinema?domain=SQS&scene_uuid=xxx-xxx&nosplash"
    host, queries = parse_url(url)
    assert host == "cinema"
    assert queries == {"domain": ["SQS"],
                       "scene_uuid": ["xxx-xxx"],
                       "nosplash": ['']}

    url = ("karabo://cinema?domain=SQS&scene_uuid=xxx-xxx&username=operator"
           "&host=localhost&port=44444")
    host, queries = parse_url(url)
    assert host == "cinema"
    assert queries == {"domain": ["SQS"],
                       "scene_uuid": ["xxx-xxx"],
                       "username": ["operator"],
                       "host": ["localhost"],
                       "port": ["44444"]}

    url = ("karabo://theatre?scene_id=COMPONENT%2FTYPE%2FMEMBER%7CSCENEname"
           "&username=operator&host=localhost&port=44444")
    host, queries = parse_url(url)
    assert host == "theatre"
    assert queries == {"scene_id": ["COMPONENT/TYPE/MEMBER|SCENEname"],
                       "username": ["operator"],
                       "host": ["localhost"],
                       "port": ["44444"]}

    url = ("karabo://theatre?scene_id=COMPONENT%2FTYPE%2FMEMBER%7CSCENEname")
    host, queries = parse_url(url)
    assert host == "theatre"
    assert queries == {"scene_id": ["COMPONENT/TYPE/MEMBER|SCENEname"]}

    url = ("karabo://theatre?scene_id=COMPONENT%2FTYPE%2FMEMBER%7CSCENEname"
           "&scene_id=COMPONENT%2FTYPE%2FMEMBER%7CSCENEname"
           "&username=operator&host=localhost&port=44444")
    host, queries = parse_url(url)
    assert host == "theatre"
    assert queries == {"scene_id": ["COMPONENT/TYPE/MEMBER|SCENEname",
                                    "COMPONENT/TYPE/MEMBER|SCENEname"],
                       "username": ["operator"],
                       "host": ["localhost"],
                       "port": ["44444"]}

    url = ("karabo://theatre?scene_id=COMPONENT%2FTYPE%2FMEMBER%7CSCENEname")
    host, queries = parse_url(url)
    assert host == "theatre"
    assert queries == {"scene_id": ["COMPONENT/TYPE/MEMBER|SCENEname"]}


def test_get_cinema_args():
    # Check query with a domain and two scene UUIDs
    queries = {"domain": ["SQS"],
               "scene_uuid": ["xxx-xxx", "yyy-yyy"]}
    assert get_cinema_args(queries) == ["SQS", "xxx-xxx", "yyy-yyy"]

    # Check query with a domain, a scene UUID and a nosplash flag
    queries = {"domain": ["SQS"],
               "scene_uuid": ["xxx-xxx"],
               "nosplash": ['']}
    assert get_cinema_args(queries) == ["SQS", "xxx-xxx", "--nosplash"]

    # Check query with two domains and a scene UUID
    queries = {"domain": ["SQS", "SCS"],
               "scene_uuid": ["xxx-xxx"]}
    with pytest.raises(ValueError):
        get_cinema_args(queries)

    # Check query with one domain and no scene UUID
    queries = {"domain": ["SQS"]}
    with pytest.raises(ValueError):
        get_cinema_args(queries)

    # Check query with blank domains and a scene UUID
    queries = {"domain": [''],
               "scene_uuid": ["xxx-xxx"]}
    with pytest.raises(ValueError):
        get_cinema_args(queries)

    # Check query for optional but both supplied host and port
    queries = {"domain": ["SQS"],
               "scene_uuid": ["xxx-xxx"],
               "host": ["localhost"],
               "port": ["44444"]}
    assert get_cinema_args(queries) == ["SQS", "xxx-xxx",
                                        "--host", "localhost",
                                        "--port", "44444"]

    # Check query for optional but only host is supplied
    queries = {"domain": ["SQS"],
               "scene_uuid": ["xxx-xxx"],
               "host": ["localhost"]}
    assert get_cinema_args(queries) == ["SQS", "xxx-xxx"]

    # Check query for optional but only port is supplied
    queries = {"domain": ["SQS"],
               "scene_uuid": ["xxx-xxx"],
               "port": ["44444"]}
    assert get_cinema_args(queries) == ["SQS", "xxx-xxx"]


def test_get_theatre_args():
    queries = {"scene_id": ["COMPONENT/TYPE/MEMBER|SCENEname"],
               "host": ["localhost"],
               "port": ["44444"]}

    assert get_theatre_args(queries) == ["COMPONENT/TYPE/MEMBER|SCENEname",
                                         "--host", "localhost",
                                         "--port", "44444"]

    queries = {"scene_id": ["COMPONENT/TYPE/MEMBER|SCENEname"]}

    assert get_theatre_args(queries) == ["COMPONENT/TYPE/MEMBER|SCENEname"]

    queries = {"scene_id": ["bla", "bli"],
               "host": ["localhost"],
               "port": ["44444"]}

    assert get_theatre_args(queries) == ["bla", "bli",
                                         "--host", "localhost",
                                         "--port", "44444"]



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QDialog, QWidget

from karabo.common.api import State
from karabo.common.scenemodel.api import StatefulIconWidgetModel
from karabo.native import Configurable, String
from karabogui.indicators import STATE_COLORS
from karabogui.testing import get_class_property_proxy, set_proxy_value

from ..statefulicon import ICONS, StatefulIconWidget

ICON_NAME = "icon_bdump"


class MockSvgWidget(QWidget):
    def load(self, svg):
        self.loaded_data = svg


class MockDialog(QDialog):
    singleton = None

    def exec(self):
        # Yeah... We need to break into the widget's code
        MockDialog.singleton = self


class Object(Configurable):
    state = String(displayType="State")


@pytest.fixture
def statefulicon_setup(gui_app):
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, "state")
    yield proxy


def test_statefulicon_basics(statefulicon_setup):
    proxy = statefulicon_setup
    model = StatefulIconWidgetModel(icon_name=ICON_NAME)
    controller = StatefulIconWidget(proxy=proxy, model=model)
    controller.create(None)
    assert controller.widget is not None

    controller.destroy()
    assert controller.widget is None


def test_statefulicon_set_value(statefulicon_setup, mocker):
    target = "karabogui.controllers.display.statefulicon.QSvgWidget"
    mocker.patch(target, new=MockSvgWidget)
    proxy = statefulicon_setup
    model = StatefulIconWidgetModel(icon_name=ICON_NAME)
    controller = StatefulIconWidget(proxy=proxy, model=model)
    controller.create(None)

    states = ("CHANGING", "ACTIVE", "PASSIVE", "DISABLED", "STATIC",
              "RUNNING", "NORMAL", "ERROR", "INIT", "UNKNOWN")

    for state in states:
        set_proxy_value(proxy, "state", state)
        color = STATE_COLORS[getattr(State, state)]
        svg = controller._icon.with_color(color)
        assert (controller.widget.loaded_data ==
                bytearray(svg, encoding="UTF-8"))


def test_statefulicon_pick_icon(statefulicon_setup, mocker):
    target = "karabogui.controllers.display.statefulicon.QDialog"
    mocker.patch(target, new=MockDialog)
    proxy = statefulicon_setup
    model = StatefulIconWidgetModel()
    controller = StatefulIconWidget(proxy=proxy, model=model)
    controller.create(None)

    # XXX: Really dig around inside the brains of the dialog opened
    # by controller._show_icon_picker.
    assert MockDialog.singleton is not None
    iconlist = MockDialog.singleton.children()[0]  # QListView
    item_model = iconlist.model()  # QStandardItemModel
    item_index = item_model.index(5, 0)  # QModelIndex
    # Mimic the user double-clicking
    iconlist.doubleClicked.emit(item_index)

    # Test that picking an icon actually set it to the scene model
    icon = item_index.data(Qt.UserRole + 1)  # QIcon
    assert controller._icon is icon
    assert model.icon_name == icon.name


def test_statefulicon_pick_icon_bailout(statefulicon_setup, mocker):
    """Test the bailout of an icon dialog"""
    target = "karabogui.controllers.display.statefulicon.QDialog"
    mocker.patch(target, new=MockDialog)
    proxy = statefulicon_setup
    model = StatefulIconWidgetModel()
    controller = StatefulIconWidget(proxy=proxy, model=model)
    assert controller._icon is None
    controller.create(None)
    assert MockDialog.singleton is not None
    no_selection = ICONS["icon_default"]
    assert controller._icon == no_selection
    assert model.icon_name == no_selection.name


def test_statefulicon_pick_not_supported_icon(statefulicon_setup, mocker):
    """Test that we maintain a not supported icon"""
    target = ("karabogui.controllers.display.statefulicon."
              "StatefulIconWidget._show_icon_picker")
    picker = mocker.patch(target)
    proxy = statefulicon_setup
    model = StatefulIconWidgetModel(icon_name="NoStateFullIcon")
    controller = StatefulIconWidget(proxy=proxy, model=model)
    controller.create(None)
    # dialog was not launched since we have an icon name!
    picker.assert_not_called()
    # The icon name is not found, the widget shows default icon
    assert controller._icon == ICONS["icon_default"]
    # However, model is not changed and keep icon name
    assert model.icon_name == "NoStateFullIcon"



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy import uic
from qtpy.QtCore import QAbstractItemModel, QModelIndex, Qt, Slot
from qtpy.QtGui import QColor
from qtpy.QtWidgets import QAbstractItemView, QDialog, QHeaderView
from traits.api import Any, Bool, HasStrictTraits, Instance, Int, List, String

from karabogui.indicators import PROPERTY_READONLY_COLOR
from karabogui.singletons.api import get_config

from .utils import get_dialog_ui

HEADER_LABELS = ["Name", "Setting"]
NAME_COLUMN = 0
SETTING_COLUMN = 1


class ConfigNode(HasStrictTraits):
    # The name of our configuration item
    name = String
    # The parent node
    parent_node = Instance("ConfigNode", allow_none=True)
    # The list of children
    children = List(Instance("ConfigNode"))
    # Check whether this config node is configurable or not
    editable = Bool
    # Level whether this config node belongs to a group or child
    level = Int(0)
    # data type of the property, e.g. bool
    dtype = Any

    def get_value(self):
        return get_config()[self.name]

    def set_value(self, value):
        get_config()[self.name] = value

    def appendChild(self, item):
        self.children.append(item)

    def child(self, row):
        return self.children[row]

    def row(self):
        if self.parent_node is None:
            return 0
        return self.parent_node.children.index(self)

    def parent(self):
        return self.parent_node


class ConfigurationModel(QAbstractItemModel):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.root = ConfigNode()
        self._get_model_data()

    def columnCount(self, parent=None):
        """Reimplemented function of QAbstractItemModel
        """
        return len(HEADER_LABELS)

    def headerData(self, column, orientation, role=Qt.DisplayRole):
        """Reimplemented function of QAbstractItemModel
        """
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return HEADER_LABELS[column]

    def index(self, row, column, parent=QModelIndex()):
        """Reimplemented function of QAbstractItemModel
        """
        if not self.hasIndex(row, column, parent):
            return QModelIndex()

        if not parent.isValid():
            parent_node = self.root
        else:
            parent_node = parent.internalPointer()

        child_node = parent_node.child(row)
        if child_node is not None:
            return self.createIndex(row, column, child_node)

    def parent(self, index):
        """Reimplemented function of QAbstractItemModel
        """
        if not index.isValid():
            return QModelIndex()

        child_node = index.internalPointer()
        if not child_node:
            return QModelIndex()

        parent_node = child_node.parent()

        if parent_node == self.root:
            return QModelIndex()

        return self.createIndex(parent_node.row(), 0, parent_node)

    def flags(self, index):
        """Reimplemented function of QAbstractItemModel"""
        if not index.isValid():
            return Qt.NoItemFlags
        if index.column() != 1:
            return Qt.ItemIsEnabled

        node = index.internalPointer()
        if node is None:
            return Qt.NoItemFlags

        flags = Qt.ItemIsEnabled | Qt.ItemIsSelectable
        if node.dtype == bool and node.editable:
            flags |= Qt.ItemIsUserCheckable
        elif node.editable:
            flags |= Qt.ItemIsEditable
        return flags

    def rowCount(self, parent=QModelIndex()):
        """Reimplemented function of QAbstractItemModel
        """
        if parent.column() > 0:
            return 0
        if not parent.isValid():
            parent = self.root
        else:
            parent = parent.internalPointer()
        return len(parent.children)

    def _get_model_data(self):
        """Setup the model data from the configuration singleton
        """
        config = get_config()
        groups = config.groups()
        for group_name in groups.keys():
            group_node = ConfigNode(name=group_name,
                                    parent_node=self.root)
            self.root.appendChild(group_node)
            for item in groups[group_name]:
                attr_item = ConfigNode(parent_node=group_node,
                                       **item.toDict(), level=1)
                group_node.appendChild(attr_item)

    def setData(self, index, value, role=Qt.EditRole):
        """Reimplemented function of QAbstractItemModel

        Active roles are CheckStateRole, DisplayRole and EditRole
        """
        if value is None:
            return False

        # Get the index's stored object
        node = index.internalPointer()
        if node is None:
            return False

        if role == Qt.CheckStateRole and node.dtype == bool:
            value = value == Qt.Checked
            node.set_value(value)
            self.dataChanged.emit(index, index)
            return True

        elif role in (Qt.DisplayRole, Qt.EditRole):
            node.set_value(value)
            self.dataChanged.emit(index, index)
            return True

        return False

    def data(self, index, role=Qt.DisplayRole):
        """Reimplemented function of QAbstractItemModel"""
        if not index.isValid():
            return None

        node = index.internalPointer()
        column = index.column()
        if column == NAME_COLUMN:
            if role == Qt.DisplayRole:
                return node.name
        elif column == SETTING_COLUMN:
            if node.level == 1:
                data = node.get_value()
                if role == Qt.CheckStateRole:
                    if node.dtype == bool:
                        return Qt.Checked if data else Qt.Unchecked
                elif role == Qt.DisplayRole:
                    return str(data)
                elif role == Qt.ForegroundRole and not node.editable:
                    return QColor(*PROPERTY_READONLY_COLOR)


class ApplicationConfigurationDialog(QDialog):
    """The dialog to view the configuration singleton
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setAttribute(Qt.WA_DeleteOnClose)
        # Don't block the event loop!
        self.setModal(False)
        filepath = get_dialog_ui("configuration.ui")
        uic.loadUi(filepath, self)

        self.tree_view.setSelectionBehavior(QAbstractItemView.SelectRows)
        item_model = ConfigurationModel(parent=self)
        self.tree_view.setModel(item_model)
        self.tree_view.setAlternatingRowColors(True)
        self.configure_header()
        self.expanded = False

    def configure_header(self):
        """Configure the treeview header to automatically resize its contents
        """
        header = self.tree_view.header()
        header.setSectionResizeMode(0, QHeaderView.ResizeToContents)
        header.setSectionResizeMode(1, QHeaderView.ResizeToContents)
        header.sectionDoubleClicked.connect(self.onDoubleClickHeader)

    @Slot()
    def onDoubleClickHeader(self):
        if self.expanded:
            self.collapseAll()
        else:
            self.expandAll()

    def collapseAll(self):
        self.expanded = False
        self.tree_view.collapseAll()

    def expandAll(self):
        self.expanded = True
        self.tree_view.expandAll()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest
from qtpy.QtWidgets import QWidget

from karabo.common.scenemodel.api import ErrorBoolModel
from karabo.native import Bool, Configurable, Hash
from karabogui.binding.api import apply_configuration
from karabogui.testing import get_class_property_proxy

from ..errorbool import ERROR_BOOL, OK_BOOL, DisplayErrorBool


class MockQSvgWidget(QWidget):
    def load(self, svg):
        self.loaded_data = svg


class Object(Configurable):
    prop = Bool(defaultValue=True)
    prop2 = Bool(defaultValue=False)


@pytest.fixture()
def controller(gui_app, mocker):
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, "prop")

    model = ErrorBoolModel()
    target = "karabogui.controllers.display.errorbool.SvgWidget"
    mocker.patch(target, new=MockQSvgWidget)
    proxy.value = False

    controller = DisplayErrorBool(proxy=proxy, model=model)
    controller.create(None)
    assert controller.widget
    yield controller
    controller.destroy()
    assert controller.widget is None


def test_errorbool_controller(controller):
    # exercise code paths
    proxy = controller.proxy
    apply_configuration(Hash("prop", True), proxy.root_proxy.binding)
    assert controller.widget.loaded_data is OK_BOOL
    apply_configuration(Hash("prop", False), proxy.root_proxy.binding)
    assert controller.widget.loaded_data is ERROR_BOOL
    assert controller.widget.toolTip() == "prop\nFalse"

    controller.model.invert = True
    assert controller.widget.loaded_data is OK_BOOL
    assert controller.widget.toolTip() == "prop\nInverted: False"



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest import mock

from qtpy.QtCore import Qt
from qtpy.QtGui import QKeyEvent
from qtpy.QtWidgets import QApplication

from karabo.common.project.api import ProjectModel
from karabo.common.scenemodel.api import SceneModel
from karabogui.testing import GuiTestCase
from karabogui.util import SignalBlocker

from ..controllers import ConfigureController, SelectScenesController
from ..wizard import CinemaWizardController


class NetworkMock:

    def __init__(self, host, port):
        self.hostname = host
        self.port = port

    def connectToServer(self):
        """Do nothing!"""


MOCKED_NETWORK = NetworkMock(host="not-localhost", port="12345")
GET_NETWORK_PATH = "karabogui.wizards.cinema.wizard.get_network"


class BaseCinemaInstallerTest(GuiTestCase):

    def setUp(self):
        super().setUp()
        self.installer = CinemaWizardController()

    def tearDown(self):
        super().tearDown()
        self.installer = None

    def assert_select_scenes_page(self, available, selected):
        current_page = self.installer.current_controller.page

        # Assert available models
        items = self.get_listwidget_items(current_page.available_listwidget)
        avail_models = [item.data(Qt.UserRole) for item in items]
        assert avail_models == available

        current_controller = self.installer.current_controller
        assert current_controller.available_scenes == available

        # Assert selected models
        items = self.get_listwidget_items(current_page.selected_listwidget)
        sel_models = [item.data(Qt.UserRole) for item in items]
        assert sel_models == selected

        assert current_controller.selected_scenes == selected
        # Assert that the wizard can go next if there are selected items
        assert current_page.isComplete() == bool(selected)

    def assert_configure_link_page(self, **kwargs):
        # Defaults here that are updated
        config = dict(host=MOCKED_NETWORK.hostname,
                      port=MOCKED_NETWORK.port,
                      show_splash=True,
                      include_host=False)
        config.update(kwargs)

        current_controller = self.installer.current_controller
        # Assert shortcut config
        for k, v in config.items():
            assert current_controller.link[k] == v

        # Assert widget values
        current_page = current_controller.page
        assert current_page.splash_checkbox.isChecked() == \
               config["show_splash"]
        assert current_page.include_host_checkbox.isChecked() == \
               config["include_host"]
        assert current_page.host_lineedit.text() == config["host"]
        assert current_page.port_lineedit.text() == config["port"]

    def start_installer(self, **traits):
        self.installer.trait_set(**traits)
        self.installer.run()

    def get_listwidget_items(self, list_widget):
        return [list_widget.item(row)
                for row in range(list_widget.count())]


@mock.patch(GET_NETWORK_PATH, return_value=MOCKED_NETWORK)
class TestSelectScenes(BaseCinemaInstallerTest):

    def test_init(self, mocked_network):
        """Test the initialization of the Select Scenes Controller"""
        available_scenes = [SceneModel(simple_name="available_1"),
                            SceneModel(simple_name="available_2")]
        selected_scenes = [SceneModel(simple_name="selected_1"),
                           SceneModel(simple_name="selected_2")]
        project_model = ProjectModel(scenes=available_scenes + selected_scenes)

        # Start the installer with a project model and a selected scene.
        self.start_installer(project_model=project_model)
        # The wizard goes directly to the select scenes page
        current_controller = self.installer.current_controller
        assert isinstance(current_controller, SelectScenesController)

    def test_changes(self, mocked_network):
        """Test the changes (add/remove) of the Select Scenes Controller"""
        available_scenes = [SceneModel(simple_name="available_1"),
                            SceneModel(simple_name="available_2")]
        selected_scenes = [SceneModel(simple_name="selected_1"),
                           SceneModel(simple_name="selected_2")]
        project_model = ProjectModel(scenes=available_scenes + selected_scenes)

        # Start the installer with a project model and a selected scene.
        self.start_installer(project_model=project_model)
        self.assert_select_scenes_page(available_scenes + selected_scenes,
                                       selected=[])

        # Add scenes
        self._add_model(selected_scenes[0])
        self.assert_select_scenes_page(available_scenes + [selected_scenes[1]],
                                       selected=[selected_scenes[0]])
        self._add_model(selected_scenes[1])
        self.assert_select_scenes_page(available_scenes, selected_scenes)

        # Remove scenes
        self._remove_model(selected_scenes[0])
        self.assert_select_scenes_page(available_scenes + [selected_scenes[0]],
                                       selected=[selected_scenes[1]])
        self._remove_model(selected_scenes[1])
        self.assert_select_scenes_page(available_scenes + selected_scenes,
                                       selected=[])

    def test_next(self, mocked_network):
        """Test the `next` of the Select Scenes Controller"""
        available_scenes = [SceneModel(simple_name="available_1"),
                            SceneModel(simple_name="available_2")]
        selected_scenes = [SceneModel(simple_name="selected_1"),
                           SceneModel(simple_name="selected_2")]
        project_model = ProjectModel(scenes=available_scenes + selected_scenes)

        # Start the installer with a project model and select the items.
        self.start_installer(project_model=project_model)
        for scene in selected_scenes:
            self._add_model(scene)

        # Go next!
        self.installer.widget.next()
        current_controller = self.installer.current_controller
        assert isinstance(current_controller, ConfigureController)

        # Go back and check that nothing has changed
        self.installer.widget.back()
        current_controller = self.installer.current_controller
        assert isinstance(current_controller, SelectScenesController)
        self.assert_select_scenes_page(available_scenes, selected_scenes)

    def _add_model(self, model):
        widget = self.installer.current_controller.page
        self._select_item(model, widget.available_listwidget)
        widget.add_button.click()

    def _remove_model(self, model):
        widget = self.installer.current_controller.page
        self._select_item(model, widget.selected_listwidget)
        widget.remove_button.click()

    def _select_item(self, model, list_widget):
        for item in self.get_listwidget_items(list_widget):
            if item.data(Qt.UserRole) is model:
                return list_widget.setCurrentItem(item)


@mock.patch(GET_NETWORK_PATH, return_value=MOCKED_NETWORK)
class TestConfigureLink(BaseCinemaInstallerTest):

    def test_init_single_selected(self, mocked_network):
        """Test the initialization of ConfigureLinkController"""
        selected_scene = SceneModel(simple_name="selected")
        available_scene = SceneModel(simple_name="available")
        project_model = ProjectModel(scenes=[available_scene, selected_scene])

        # Start the installer with a project model and a selected scene.
        self.start_installer(selected_scenes=[selected_scene],
                             project_model=project_model)
        # The wizard goes directly to the configure shortcut page

        current_controller = self.installer.current_controller
        assert isinstance(current_controller, ConfigureController)
        self.assert_configure_link_page(uuids=[selected_scene.uuid])
        current_page = current_controller.page
        assert current_page.isComplete()

    def test_init_multiple_selected(self, mocked_network):
        """Test the setup of ConfigureLinkController with multiple scenes"""

        available_scenes = [SceneModel(simple_name="available_1"),
                            SceneModel(simple_name="available_2")]
        selected_scenes = [SceneModel(simple_name="selected_1"),
                           SceneModel(simple_name="selected_2")]
        project_model = ProjectModel(simple_name="Project",
                                     scenes=available_scenes + selected_scenes)

        # Start the installer with a project model and a selected scene.
        self.start_installer(selected_scenes=selected_scenes,
                             project_model=project_model)

        # The wizard goes directly to the configure shortcut page
        current_controller = self.installer.current_controller
        assert isinstance(current_controller, ConfigureController)
        self.assert_configure_link_page(
            uuids=[scene.uuid for scene in selected_scenes])
        current_page = current_controller.page
        assert current_page.isComplete()

    def test_changes(self, mocked_network):
        """Test the changes of the link configuration and validation"""
        selected_scene = SceneModel(simple_name="selected")
        available_scene = SceneModel(simple_name="available")
        project_model = ProjectModel(scenes=[available_scene, selected_scene])

        # Start the installer with a project model and a selected scene.
        self.start_installer(selected_scenes=[selected_scene],
                             project_model=project_model)

        page = self.installer.current_controller.page
        link = self.installer.current_controller.link

        # Toggle splash checkbox
        page.splash_checkbox.toggle()
        assert link["show_splash"] == page.splash_checkbox.isChecked()

        # Toggle login checkbox
        page.include_host_checkbox.toggle()
        is_checked = page.include_host_checkbox.isChecked()
        assert link["include_host"] == is_checked
        assert page.host_lineedit.isVisible() == is_checked
        assert page.port_lineedit.isVisible() == is_checked

        # Edit host field: invalid
        self._edit_lineedit('', page.host_lineedit)
        assert not page.isComplete()

        # Edit host field: valid
        self._edit_lineedit("some-host", page.host_lineedit)
        assert page.isComplete()
        assert link["host"] == "some-host"

        # Edit port field: invalid
        self._edit_lineedit('', page.port_lineedit)
        assert not page.isComplete()

        # Edit port: valid
        self._edit_lineedit("34567", page.port_lineedit)
        assert page.isComplete()
        assert link["port"] == "34567"

    def _edit_lineedit(self, text, line_edit):
        line_edit.setFocus()
        line_edit.setText(text)
        key_event = QKeyEvent(QKeyEvent.KeyPress, Qt.Key_Enter, Qt.NoModifier)
        with SignalBlocker(self.installer.widget):
            QApplication.sendEvent(line_edit, key_event)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from datetime import UTC, datetime
from platform import system

import pytest
from qtpy.QtCore import Qt

from karabo.common.scenemodel.api import CurveType, extract_graph_curve_option
from karabo.native import Bool, Configurable, Float, Int32
from karabogui.controllers.trendmodel import UPTIME
from karabogui.graph.plots.utils import create_curve_options
from karabogui.testing import get_class_property_proxy, set_proxy_value
from karabogui.util import process_qt_events

from ..time_graphs import DisplayTrendGraph
from .data import build_historic_data_float


class Object(Configurable):
    prop = Bool(defaultValue=True)
    fprop = Float(defaultValue=1.0)
    iprop = Int32(defaultValue=2.1)


def test_trendline(gui_app):
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, "prop")
    controller = DisplayTrendGraph(proxy=proxy)
    controller.create(None)

    set_proxy_value(proxy, "prop", True)
    process_qt_events()

    assert controller._x_detail == UPTIME

    # actions
    actions = controller.widget.actions()
    texts = [action.text() for action in actions]
    assert len(actions) == 8
    assert "Range Y" in texts
    assert "View" in texts

    # historic data
    data = build_historic_data_float()
    t0 = data[0]["v", ...]["sec"]
    t1 = data[-1]["v", ...]["sec"]

    controller.set_time_interval(t0, t1)

    proxy.binding.historic_data = data
    process_qt_events()

    # Range settings
    view_box = controller._plot.plotItem.vb
    view_box.setXRange(1570536895, 1571536895, padding=0.05)

    dt_start = controller.widget.dt_start.dateTime().toUTC()
    dt_end = controller.widget.dt_end.dateTime().toUTC()

    assert dt_start.toString(Qt.ISODate) == "2019-10-07T22:21:35Z"
    assert dt_end.toString(Qt.ISODate) == "2019-10-20T15:54:55Z"

    # Additional proxies
    assert len(controller.proxies) == 1

    float_proxy = get_class_property_proxy(schema, "fprop")
    controller.visualize_additional_property(float_proxy)
    assert len(controller.proxies) == 2

    int_proxy = get_class_property_proxy(schema, "iprop")
    controller.visualize_additional_property(int_proxy)
    assert len(controller.proxies) == 3

    # and now add the curve options and alter them
    proxy_curve = controller._curves[proxy]
    assert proxy_curve.name() == "prop"
    assert proxy_curve.opts["pen"].color().name() == "#3399ff"

    options = create_curve_options(
        controller._curves, {}, CurveType.Trend)
    options[proxy_curve]["name"] = "New legend name"
    options[proxy_curve]["pen_color"] = "#ff3333"

    controller.set_curve_options(options)
    assert len(controller.model.curve_options) == 3

    # Verify the changed settings
    assert controller._curves[proxy].name() == "New legend name"
    assert proxy_curve.opts["pen"].color().name() == "#ff3333"

    # Check all 3
    for p in [int_proxy, float_proxy, proxy]:
        curve_opt = extract_graph_curve_option(controller.model, p.key)
        assert len(curve_opt), "Curve options cannot be empty"

    # Float removal
    controller.remove_additional_property(float_proxy)
    assert len(controller.proxies) == 2
    curve_opt = extract_graph_curve_option(controller.model, float_proxy.key)
    assert not len(curve_opt), "Curve options must be empty"

    # Integer removal
    controller.remove_additional_property(int_proxy)
    assert len(controller.proxies) == 1
    curve_opt = extract_graph_curve_option(controller.model, int_proxy.key)
    assert not len(curve_opt), "Curve options must be empty"

    # reset remaining
    controller.reset_curve_options()
    assert controller._curves[proxy].name() == "prop"
    assert proxy_curve.opts["pen"].color().name() == "#3399ff"
    assert len(controller.model.curve_options) == 0

    # Destroy controller
    controller.destroy()
    assert controller.widget is None


@pytest.mark.skipif(system() == "Windows",
                    reason="This tests is Unix specific")
def test_time_graph_range_update(gui_app):
    """When the viewbox change, we must make sure the time axis is
    not overflown"""
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, "prop")
    controller = DisplayTrendGraph(proxy=proxy)
    controller.create(None)

    plot_view = controller._plot
    plot_item = plot_view.plotItem

    view_box = plot_item.vb
    view_box.setXRange(datetime(1378, 12, 31).timestamp(),
                       datetime(2070, 12, 31).timestamp(), update=False)
    process_qt_events()
    # Relax the assertion as the CI is non deterministic
    x_min, x_max = plot_item.getAxis("bottom").range
    assert datetime.fromtimestamp(x_min, UTC).year == 1970
    assert datetime.fromtimestamp(x_max, UTC).year == 2038

    controller.destroy()
    assert controller.widget is None



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from pathlib import Path

import numpy as np
from qtpy.QtGui import QValidator
from qtpy.QtWidgets import QSlider, QSpinBox

from karabogui.util import (
    InputValidator, SignalBlocker, _get_invalid_chars, convert_npy_to_csv,
    create_list_string, create_table_string, qtversion_compatible,
    version_compatible)


def test_assert_filename():
    def assert_filename(filename, *, invalid):
        invalid_chars = _get_invalid_chars(filename)
        assert set(list(invalid)) == set(invalid_chars)

    assert_filename("foo", invalid='')
    assert_filename("FoO/Bar-bAz_123", invalid='')
    assert_filename("foo bar", invalid=' ')
    assert_filename("foo!@#$%&*()+<>@:", invalid='!@#$%&*()+<>:')


def test_input_validator(gui_app):
    validator = InputValidator()
    valid, _, _ = validator.validate("1", 0)
    assert valid == QValidator.Acceptable

    valid, _, _ = validator.validate("w23-", 0)
    assert valid == QValidator.Acceptable

    valid, _, _ = validator.validate("w23-w+", 0)
    assert valid == QValidator.Invalid

    valid, _, _ = validator.validate("1+", 0)
    assert valid == QValidator.Invalid

    valid, _, _ = validator.validate("1#", 0)
    assert valid == QValidator.Invalid

    valid, _, _ = validator.validate("1|", 0)
    assert valid == QValidator.Invalid

    valid, _, _ = validator.validate("+a", 0)
    assert valid == QValidator.Invalid

    valid, _, _ = validator.validate("-a", 0)
    assert valid == QValidator.Acceptable

    # Macro case
    validator = InputValidator("macro")
    valid, _, _ = validator.validate("1", 0)
    assert valid == QValidator.Invalid

    valid, _, _ = validator.validate("-a", 0)
    assert valid == QValidator.Invalid

    valid, _, _ = validator.validate("+a", 0)
    assert valid == QValidator.Invalid

    valid, _, _ = validator.validate("w23", 0)
    assert valid == QValidator.Acceptable


def test_signal_blocker(gui_app):
    count = 0

    def slotReceived(*args, **kwargs):
        nonlocal count
        count += 1

    spinbox = QSpinBox()
    spinbox.valueChanged.connect(slotReceived)
    assert count == 0
    spinbox.setValue(100)
    assert count == 1

    # Update the spinbox, but block the signals
    with SignalBlocker(spinbox):
        for i in range(10):
            spinbox.setValue(i)
        assert spinbox.value() == i

    # Still only once
    assert count == 1
    spinbox.setValue(100)
    assert count == 2

    # Reset the counter and setup a slider
    count = 0

    slider = QSlider()
    slider.valueChanged.connect(slotReceived)
    slider.setRange(0, 100)
    slider.setValue(20)
    assert count == 1

    # Works, now block both
    with SignalBlocker(spinbox, slider):
        for i in range(10):
            spinbox.setValue(i)
            slider.setValue(i)
        assert spinbox.value() == i
        assert slider.value() == i
    assert count == 1

    # Other initialization
    count = 0
    blocker = SignalBlocker(spinbox)
    assert len(blocker.objects) == 1
    with blocker:
        spinbox.setValue(10000)
    assert count == 0


def test_version_compatible():
    """Test that we can verify versions"""
    assert version_compatible("2.14.0a2", 2, 14)
    assert version_compatible("2.14.0a2", 2, 13)
    assert version_compatible("2.14.0rc1", 2, 14)
    assert not version_compatible("2.14.0a2", 2, 15)
    assert version_compatible("asdeaeddevelopmode", 2, 14)
    assert version_compatible("asdeaeddevelopmode", 12, 33333)
    # minor smaller
    assert version_compatible("3.1.0a2", 2, 15)

    # Test garbage
    assert not version_compatible("2.&7as7.13", 2, 0)


def test_qversion_compatible():
    assert qtversion_compatible(5, 13)
    assert qtversion_compatible(5, 9)


def test_html_list_string():
    info = ["1", "2", 4, []]
    html = create_list_string(info)
    assert html == "<ul><li>1</li><li>2</li><li>4</li><li>[]</li></ul>"


def test_html_table_string():
    info = {"1": "a", 2: 3}
    html = create_table_string(info)
    assert html == "<table>" \
                   "<tr><td><b>1</b>:   </td><td>a</td></tr>" \
                   "<tr><td><b>2</b>:   </td><td>3</td></tr>" \
                   "</table>"


def test_npy2csv():
    """Test conversion of numpy files - npy and npz- to csv file."""
    # npy file.
    TEST_DATA_DIR = Path(__file__).resolve().parent / "data"
    input_file = TEST_DATA_DIR / "single_array.npy"
    assert input_file.exists()
    convert_npy_to_csv(input_file)
    output_file = TEST_DATA_DIR / "single_array.csv"

    x = np.arange(6.0000, dtype=float)
    y = np.arange(1.0, 12.0, step=2.0, dtype=float)
    expected = np.vstack((x, y))

    data_from_csv = np.genfromtxt(output_file, delimiter=',')
    np.testing.assert_array_equal(data_from_csv, expected)
    Path(output_file).unlink()

    # npz file.
    input_file = TEST_DATA_DIR / "multiple_arrays.npz"
    assert input_file.exists()
    convert_npy_to_csv(input_file)
    output_file = TEST_DATA_DIR / "multiple_arrays.csv"
    data_from_csv = np.genfromtxt(output_file, delimiter=',')

    x = np.arange(15.0, dtype=float)
    y1 = np.arange(10.0, 151.0, step=10.0, dtype=float)
    y2 = np.full(15, 10.0)
    y3 = np.arange(15.0, dtype=float)
    plot1 = np.vstack((x, y1))
    plot2 = np.vstack((x, y2))
    plot3 = np.vstack((x, y3))
    expected = np.vstack((plot1, plot2, plot3))

    np.testing.assert_array_equal(data_from_csv, expected)
    Path(output_file).unlink()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest

from karabo.native import Configurable, Float, String
from karabogui.binding.api import (
    DeviceClassProxy, PropertyProxy, ProxyStatus, build_binding)
from karabogui.indicators import ALL_OK_COLOR
from karabogui.testing import set_proxy_value

from ..label import DisplayLabel


class Object(Configurable):
    string = String()
    alarms = Float()
    faulty = Float(displayType="fmt|{:*1271f}")


@pytest.fixture
def label_setup(gui_app):
    schema = Object.getClassSchema()
    binding = build_binding(schema)
    device = DeviceClassProxy(binding=binding, server_id="Fake",
                              status=ProxyStatus.OFFLINE)
    string = PropertyProxy(root_proxy=device, path="string")
    alarms = PropertyProxy(root_proxy=device, path="alarms")
    faulty = PropertyProxy(root_proxy=device, path="faulty")
    yield string, alarms, faulty


def test_basics(label_setup):
    string, alarms, faulty = label_setup
    controller = DisplayLabel(proxy=string)
    controller.create(None)
    assert controller.widget is not None

    controller.destroy()
    assert controller.widget is None


def test_set_string_value(label_setup):
    string, alarms, faulty = label_setup
    controller = DisplayLabel(proxy=string)
    controller.create(None)
    set_proxy_value(string, "string", "hello")
    assert controller.internal_widget.text() == "hello"


def test_alarm_color(label_setup):
    string, alarms, faulty = label_setup
    controller = DisplayLabel(proxy=alarms)
    controller.create(None)
    set_proxy_value(alarms, "alarms", 0.75)
    assert controller.internal_widget.text() == "0.75"
    assert controller.bg_color == ALL_OK_COLOR
    set_proxy_value(alarms, "alarms", 3.0)
    assert controller.bg_color == ALL_OK_COLOR
    set_proxy_value(alarms, "alarms", 1.5)
    assert controller.bg_color == ALL_OK_COLOR


def test_wrong_format(label_setup):
    string, alarms, faulty = label_setup
    controller = DisplayLabel(proxy=faulty)
    controller.create(None)
    set_proxy_value(faulty, "faulty", 0.25)
    assert controller.internal_widget.text() == "0.25"



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest
from numpy.testing import assert_array_equal
from pyqtgraph import ColorMap
from qtpy.QtWidgets import QGraphicsTextItem

from karabo.common.scenemodel.api import (
    CrossROIData, ImageGraphModel, RectROIData)
from karabo.native import Encoding
from karabogui.binding.builder import build_binding
from karabogui.binding.config import apply_configuration
from karabogui.binding.proxy import DeviceProxy, PropertyProxy
from karabogui.controllers.display.tests.image import (
    get_image_hash, get_pipeline_schema)
from karabogui.graph.common.api import COLORMAPS
from karabogui.graph.common.enums import AuxPlots, MouseTool, ROITool

from ..image_graph import DisplayImageGraph


@pytest.fixture()
def image_graph_setup(gui_app, mocker):
    schema = get_pipeline_schema(stack_axis=False)
    binding = build_binding(schema)
    root_proxy = DeviceProxy(binding=binding)

    output_proxy = PropertyProxy(root_proxy=root_proxy,
                                 path="output.data")
    img_proxy = PropertyProxy(root_proxy=root_proxy,
                              path="output.data.image")
    controller = DisplayImageGraph(proxy=img_proxy)

    mocker.patch.object(QGraphicsTextItem, "setHtml")
    controller.create(None)

    yield controller, output_proxy, img_proxy

    controller.destroy()
    assert controller.widget is None


def _click_toolbar(gui_app, controller, tool):
    toolbar = controller.widget.toolbar
    toolbar.toolsets[tool.__class__].buttons[tool].click()
    # give some process event time
    gui_app.processEvents()


def _assert_gray_features(controller, enabled):
    widget = controller.widget
    cmap = widget.configuration.get("colormap", "none")
    lut = _calc_cmap_lut(cmap)

    image_item = widget.plotItem.imageItem
    ndim = 2 if enabled else 3
    assert image_item.image.ndim == ndim
    assert_array_equal(image_item.lut, lut)

    if enabled:
        assert widget._colorbar is not None
        assert_array_equal(widget._colorbar.barItem.lut, lut)

    else:
        assert widget._colorbar is None

    expected_plot = widget.configuration.get("aux_plots", AuxPlots.NoPlot)
    if not enabled:
        expected_plot = AuxPlots.NoPlot
    assert widget._aux_plots.current_plot == expected_plot

    buttons = widget.toolbar.toolsets.get(AuxPlots).buttons
    for tool, button in buttons.items():
        assert button.isEnabled() == enabled
        assert button.isChecked() == (tool == expected_plot)


def _calc_cmap_lut(cmap):
    return ColorMap(*zip(*COLORMAPS[cmap])).getLookupTable(alpha=False)


def test_toolbar_apply(gui_app, image_graph_setup):
    """The toolbar apply in the image graph"""
    controller, output_proxy, img_proxy = image_graph_setup
    image_hash = get_image_hash()
    apply_configuration(image_hash, output_proxy.binding)

    # Assert current state
    model = controller.model
    assert model.aux_plots == AuxPlots.NoPlot
    assert len(model.roi_items) == 0

    # Toolbar configuration
    _click_toolbar(gui_app, controller, AuxPlots.ProfilePlot)
    _click_toolbar(gui_app, controller, MouseTool.Picker)

    # Add some ROIs
    roi_controller = controller.widget.roi
    roi_controller.add(ROITool.Crosshair, (50, 50), name="Cross 1")
    roi_controller.add(ROITool.Crosshair, (35, 35), name="Cross 2")
    roi_controller.add(ROITool.Rect, (15, 15), (10, 10), name="Rect")

    expected_configs = {
        ROITool.Crosshair: [("Cross 1", (50, 50)), ("Cross 2", (35, 35))],
        ROITool.Rect: [("Rect", (15, 15, 10, 10))]}

    # Assert the expected configuration
    for tool, roi_items in roi_controller._rois.items():
        configs = expected_configs[tool]
        for (name, coords), roi_item in zip(configs, roi_items):
            assert roi_item.name == name
            assert roi_item.coords == coords

    apply_action = controller.widget.actions()[2]
    assert apply_action.text() == "Set ROI and Aux"

    # Save to model
    apply_action.triggered.emit(True)
    # give some process event time
    gui_app.processEvents()

    model = controller.model
    assert model.aux_plots == AuxPlots.ProfilePlot
    assert len(model.roi_items) == 3


def test_restore_widget(image_graph_setup, mocker):
    """Create a model with ROI settings and restore in Image Graph"""
    _, _, img_proxy = image_graph_setup
    model = ImageGraphModel()
    model.aux_plots = AuxPlots.ProfilePlot

    roi_items = [CrossROIData(**{"roi_type": ROITool.Crosshair,
                                 "x": 15, "y": 15, "name": "CrossROI"}),
                 RectROIData(**{"roi_type": ROITool.Rect,
                                "x": 25, "y": 25,
                                "w": 10, "h": 10,
                                "name": "RectROI"})]
    model.roi_items = roi_items

    controller = DisplayImageGraph(proxy=img_proxy, model=model)
    mocker.patch.object(QGraphicsTextItem, "setHtml")
    controller.create(None)

    # Assert the information was loaded
    widget = controller.widget

    # Assert ROI configuration
    expected_roi_config = {ROITool.Crosshair: ("CrossROI", (15, 15)),
                           ROITool.Rect: ("RectROI", (25, 25, 10, 10))}

    for tool, roi_items in widget.roi.roi_items.items():
        roi_item = roi_items[0]
        name, coords = expected_roi_config[tool]
        assert roi_item.coords == coords
        assert roi_item.name == name

    menu = widget.plotItem.vb.menu
    assert len(menu.actions()) == 3
    assert menu.actions()[0].text() == "View all"
    assert menu.actions()[1].text() == ""
    assert menu.actions()[2].text() == "Undock"

    path = "karabogui.controllers.display.image_graph.broadcast_event"
    broadcast = mocker.patch(path)
    menu.actions()[2].trigger()
    broadcast.assert_called_once()


def test_value_update(image_graph_setup):
    # Restore aux plots
    controller, output_proxy, _ = image_graph_setup
    widget = controller.widget
    widget.restore({"aux_plots": AuxPlots.ProfilePlot,
                    "colormap": "viridis"})
    assert widget._aux_plots.current_plot == AuxPlots.ProfilePlot

    # Test update with dim=3, encoding=GRAY
    image_hash = get_image_hash(dimZ=3, stack_axis=False)
    apply_configuration(image_hash, output_proxy.binding)
    _assert_gray_features(controller, enabled=True)

    # Test update with dim=3, encoding=RGB
    image_hash = get_image_hash(dimZ=3, encoding=Encoding.RGB,
                                stack_axis=False)
    apply_configuration(image_hash, output_proxy.binding)
    _assert_gray_features(controller, enabled=False)

    # Now, we update with dim=3, encoding=GRAY again
    image_hash = get_image_hash(dimZ=3, stack_axis=False)
    apply_configuration(image_hash, output_proxy.binding)
    _assert_gray_features(controller, enabled=True)
    assert controller._plot.imageItem.image.shape == (30, 40)

    # Erase widget, a default with size 10, 10 is set
    controller.clear_widget()
    assert controller._plot.imageItem.image.shape == (10, 10)


def test_undock_image(image_graph_setup, mocker):
    """Create a model with ROI settings and undock the Image Graph"""
    _, _, img_proxy = image_graph_setup
    model = ImageGraphModel()
    model.aux_plots = AuxPlots.ProfilePlot
    model.undock = True

    roi_items = [CrossROIData(**{"roi_type": ROITool.Crosshair,
                                 "x": 15, "y": 15, "name": "CrossROI"}),
                 RectROIData(**{"roi_type": ROITool.Rect,
                                "x": 25, "y": 25,
                                "w": 10, "h": 10,
                                "name": "RectROI"})]
    model.roi_items = roi_items

    controller = DisplayImageGraph(proxy=img_proxy, model=model)
    mocker.patch.object(QGraphicsTextItem, "setHtml")
    controller.create(None)

    # Assert the information was loaded
    widget = controller.widget

    # Assert ROI configuration
    expected_roi_config = {ROITool.Crosshair: ("CrossROI", (15, 15)),
                           ROITool.Rect: ("RectROI", (25, 25, 10, 10))}

    for tool, roi_items in widget.roi.roi_items.items():
        roi_item = roi_items[0]
        name, coords = expected_roi_config[tool]
        assert roi_item.coords == coords
        assert roi_item.name == name
        assert not roi_item.translatable

    menu = controller.widget.plotItem.vb.menu
    assert len(menu.actions()) == 1
    assert menu.actions()[0].text() == "View all"


def test_save_color_levels(gui_app, image_graph_setup):
    controller, output_proxy, img_proxy = image_graph_setup
    image_hash = get_image_hash()
    apply_configuration(image_hash, output_proxy.binding)

    # When no levels saved.
    model = controller.model
    assert model.color_levels == []

    # Change the levels value.
    widget = controller.widget
    levels = [10, 45]
    widget.plotItem.set_image_levels(levels)
    assert model.color_levels == []

    save_action = controller.widget.actions()[1]
    assert save_action.text() == "Save Color Levels"
    save_action.triggered.emit(True)
    gui_app.processEvents()

    model = controller.model
    assert model.color_levels == [10, 45]

    # Set auto-levels
    widget.plotItem.set_image_levels(None)
    save_action.triggered.emit(True)
    gui_app.processEvents()

    model = controller.model
    assert model.color_levels == []



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on November 7, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from collections import OrderedDict
from os import listdir
from os.path import isfile, join, splitext
from xml.etree.ElementTree import Element, parse, tostring

from karabo.common.scenemodel.const import NS_SVG


class ColorChangeIcon:
    """An SVG group element which can have fill colors changed in one or more
    subelements.
    """
    def __init__(self, element, white_filled_subelements, description):
        self.element = element
        self.name = element.get('id')
        self.white_filled_subelements = white_filled_subelements
        self.description = description

    def with_color(self, color):
        """ Return the XML for the icon with white fill colors replaced
        """

        # make sure color is in hex representation
        if isinstance(color, (tuple, list)):
            color = '#%02x%02x%02x' % color[:3]

        for element in self.white_filled_subelements:
            style = _split_style_attr(element.get('style'))
            style['fill'] = color
            element.set('style', _join_style_attr(style))

        root = Element(NS_SVG+'svg')
        root.append(self.element)
        return tostring(root, encoding='unicode')


def get_color_change_icon(path):
    """Icon from an SVG file which contain regions with a
    fill color of white.

    :param path: A path to an SVG file
    :return: A list of ``ColorChangeIcon`` instances
    """
    def _iter_children_with_styles(element):
        for child in element:
            style = child.get('style')
            if style is not None:
                yield child, style

    def _iter_icon_groups(root):
        for group in _iter_groups_r(root):
            label = group.get('id')
            if label is not None and label.startswith('icon'):
                yield group

    def _iter_groups_r(element):
        for group in element.findall(NS_SVG + 'g'):
            yield group
            yield from _iter_groups_r(group)

    tree = parse(path)

    try:
        icon = list(_iter_icon_groups(tree))[0]
    except IndexError:
        return None
    description = icon.findall(NS_SVG + 'desc')[0].text
    styleable = [(c, _split_style_attr(s))
                 for c, s in _iter_children_with_styles(icon)]
    white_subelements = [c for c, s in styleable
                         if s.get('fill', '').lower() == '#ffffff']
    return ColorChangeIcon(icon, white_subelements, description)


def get_color_change_icons(path):
    """
    Extract color change icons from SVG files located in `path`.
    """
    icons = {}
    for in_file in listdir(path):
        fpath = join(path, in_file)
        if isfile(fpath) and splitext(fpath)[-1] == '.svg':
            icon = get_color_change_icon(join(path, in_file))
            if icon is not None:
                icons[icon.name] = icon

    return icons

# ----------------------------------------------------------------------


def _join_style_attr(style):
    return ';'.join([f'{k}:{v}' for k, v in style.items()])


def _split_style_attr(style):
    return OrderedDict((k, v)
                       for k, v in (s.split(':') for s in style.split(';')))



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.\
from functools import partial

from qtpy import uic
from qtpy.QtCore import QBuffer, QByteArray, QIODevice, QSize, Qt, Signal, Slot
from qtpy.QtGui import QColor, QIcon, QPainter, QPixmap
from qtpy.QtWidgets import (
    QAction, QActionGroup, QCheckBox, QColorDialog, QComboBox, QDialog,
    QDialogButtonBox, QGraphicsPixmapItem, QGraphicsScene, QGridLayout, QLabel,
    QLineEdit, QSizePolicy, QTableWidgetItem, QToolButton, QToolTip, QWidget)

from karabo.common.scenemodel.api import create_base64image
from karabo.native import Hash, HashList
from karabogui import icons
from karabogui.const import IS_MAC_SYSTEM
from karabogui.dialogs.logbook_drawing_tools import (
    BaseDrawingTool, CropTool, get_tools)
from karabogui.dialogs.utils import get_dialog_ui
from karabogui.events import (
    KaraboEvent, register_for_broadcasts, unregister_from_broadcasts)
from karabogui.singletons.api import get_config, get_network
from karabogui.util import SignalBlocker
from karabogui.validators import NumberValidator
from karabogui.widgets.toolbar import ToolBar

MIN_ZOOM_FACTOR = 25
MAX_ZOOM_FACTOR = 250
ZOOM_INCREMENT = 1.1

MAX_ALLOWED_LOGBOOK = 3

CHECKBOX_STYLE = """
QCheckBox {
    text-align: center;
    margin-left:50%;
    margin-right:50%;
}
"""

LOGBOOK_IMAGE = "Image"
LOGBOOK_DATA = "Data"

DESTINATION_TYPE_HINT = """<b>Active Proposals.</b><br>
<i>Ready Proposals</i>.<br>
Operational Logbooks."""


class DestinationWidget(QWidget):
    """A container  with widgets to select the Stream and Topic of the
    logbook """

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self._topics = {}
        layout = QGridLayout(self)
        layout.setSpacing(1)
        layout.setVerticalSpacing(2)
        layout.setContentsMargins(0, 0, 0, 0)
        self.combo_stream = QComboBox()
        self.combo_topic = QComboBox()

        self.create_topic = QToolButton()
        self.create_topic.setIcon(icons.edit)
        self.create_topic.setCheckable(True)

        self.combo_stream.currentTextChanged.connect(self._update_topics)
        self.create_topic.clicked.connect(self.allow_topic_creation)

        stream_label = QLabel("Stream")
        stream_label.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        topic_label = QLabel("Topic")
        topic_label.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)

        info_button = QToolButton()
        info_button.setIcon(icons.info)
        info_button.setToolTip(DESTINATION_TYPE_HINT)
        info_button.clicked.connect(self._show_info_popup)
        info_button.setStyleSheet("QToolButton{border:none}")

        layout.addWidget(stream_label, 0, 0)
        layout.addWidget(info_button, 0, 2)
        layout.addWidget(self.combo_stream, 0, 1)
        layout.addWidget(topic_label, 1, 0)
        layout.addWidget(self.combo_topic, 1, 1)
        layout.addWidget(self.create_topic, 1, 2)

    def update_destinations(self, data):
        """
        Update the comboboxes with the available destinations - stream/topic.

        :param data: A HashList with available LogBooks: Hash with the
            following fields:
                "name" : Logbook identifier
                "destination": Url or folder of the logbook.
                "topics": List of topics in the stream.
        """
        with SignalBlocker(self.combo_stream):
            self.combo_stream.clear()
        self._topics.clear()

        for index, proposal in enumerate(data):
            name = proposal.get("name")
            destination = proposal.get("destination")
            status = proposal.get("beamlineStatus")
            self._topics[name] = proposal.get("topics")
            self.combo_stream.addItem(name)
            self.combo_stream.setItemData(index, destination)
            font = self.combo_stream.font()
            if status == "Active":
                font.setBold(True)
            if status == "Ready":
                font.setItalic(True)
            self.combo_stream.setItemData(index, font, Qt.FontRole)

    def select_destination(self, stream: str, topic: str):
        self.combo_stream.setCurrentText(stream)
        self.combo_topic.setCurrentText(topic)

    @Slot()
    def _show_info_popup(self):
        button = self.sender()
        QToolTip.showText(button.mapToGlobal(button.rect().topLeft()),
                          button.toolTip(), button)

    @Slot(str)
    def _update_topics(self, text):
        """Update the combobox with available topics when the stream
        selection change"""
        self.combo_topic.clear()
        topics = self._topics[text]
        self.combo_topic.addItems(topics)

    @Slot(bool)
    def allow_topic_creation(self, toggled):
        """Make the combobox editable to allow creation of new topic"""
        if toggled:
            self.combo_topic.setEditable(True)
            self.combo_topic.lineEdit().selectAll()
            self.combo_topic.setFocus(True)
            self.combo_topic.lineEdit().editingFinished.connect(
                self.add_new_topic)
        else:
            new_topic = self.combo_topic.lineEdit().text()
            self._add_topic(new_topic)

    @Slot()
    def add_new_topic(self):
        """Add a new item to the Topic combobox"""
        new_topic = self.combo_topic.lineEdit().text()
        if not new_topic.strip():
            self.combo_topic.setEditable(False)
            self.create_topic.setChecked(False)
            return
        self._add_topic(new_topic)

    def _add_topic(self, new_topic):
        if self.combo_topic.findText(new_topic) == -1:
            self.combo_topic.addItem(new_topic)
        self.combo_topic.setEditable(False)
        self.combo_topic.setCurrentText(new_topic)
        self.create_topic.setChecked(False)

    @property
    def topic(self):
        return self.combo_topic.currentText()

    @property
    def stream(self):
        return self.combo_stream.currentText()


class Canvas(QGraphicsScene):
    """
    QGraphicsScene to draw shapes using mouse to annotate the image.
    """
    resetTool = Signal()
    fitViewRequested = Signal()

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.drawing_tool = None

    def mousePressEvent(self, event):
        if self._can_draw(event):
            self.drawing_tool.mouse_down(self, event)
        else:
            self.set_drawing_tool(None)
        super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        if self._can_draw(event):
            self.drawing_tool.mouse_move(self, event)
        super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        if self._can_draw(event):
            self.drawing_tool.mouse_up(self, event)
        elif isinstance(self.drawing_tool, CropTool):
            self.removeItem(self.drawing_tool.graphics_item)
        self.set_drawing_tool(None)
        super().mouseReleaseEvent(event)

    def _can_draw(self, event):
        return (self.drawing_tool is not None and
                self.sceneRect().contains(event.scenePos()))

    def set_drawing_tool(self, tool):
        """
        Set a drawing tool. Pass None to clear the active drawing tool.
        """
        assert tool is None or isinstance(tool, BaseDrawingTool)
        self.drawing_tool = tool
        if tool is None:
            self.resetTool.emit()

    def set_drawing_color(self, color):
        if self.drawing_tool is not None:
            self.drawing_tool.set_pen_color(color)

    def crop(self, rect):
        for item in self.items():
            if isinstance(item, QGraphicsPixmapItem):
                break
            item = None
        if item is None:
            return
        rect = rect.normalized()
        height = int(rect.height())
        width = int(rect.width())
        pixmap = QPixmap(width, height)
        with QPainter(pixmap) as painter:
            self.render(painter, source=rect, mode=Qt.KeepAspectRatio)
        item.setPixmap(pixmap)
        self.setSceneRect(0, 0, rect.width(), rect.height())
        self.fitViewRequested.emit()

    @property
    def pixmap_item(self):
        for item in self.items():
            if isinstance(item, QGraphicsPixmapItem):
                return item.pixmap()


class LogBookPreview(QDialog):
    """A Dialog to preview the data to the LogBook"""

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setModal(False)
        uic.loadUi(get_dialog_ui("logbook.ui"), self)

        self.action_group_draw = {}
        self._topics = {}

        self.pixmap = parent.grab()
        self.combo_datatype.addItem(LOGBOOK_IMAGE)
        if parent.info() is not None:
            self.combo_datatype.addItem(LOGBOOK_DATA)

        self.combo_datatype.setCurrentIndex(0)
        self.combo_datatype.currentIndexChanged.connect(
            self._on_datatype_changed)

        self.ok_button = self.buttonBox.button(QDialogButtonBox.Ok)
        self.ok_button.setText("Save")
        self.ok_button.setEnabled(False)

        self.canvas = Canvas(self)
        self.canvas.addPixmap(self.pixmap)
        self.view.setScene(self.canvas)
        self.canvas.resetTool.connect(self.reset_tool)
        self.canvas.fitViewRequested.connect(self._fit_in_view)

        validator = NumberValidator(
            minInc=MIN_ZOOM_FACTOR, maxInc=MAX_ZOOM_FACTOR, decimals=1)
        self.zoom_factor_edit = QLineEdit()
        self.zoom_factor_edit.setValidator(validator)
        self.zoom_factor.setLineEdit(self.zoom_factor_edit)

        self._fit_in_view()

        self.data = None
        self._destinations = []

        self.zoom_factor.currentIndexChanged.connect(self.change_zoom_factor)
        self.zoom_factor_edit.editingFinished.connect(self.change_zoom_factor)

        self._event_map = {
            KaraboEvent.ActiveDestinations: self._event_destinations,
            KaraboEvent.NetworkConnectStatus: self._event_network
        }
        register_for_broadcasts(self._event_map)
        get_network().listDestinations()

        self.checkboxes = []
        self._load_table()
        self.select_all.clicked.connect(self._select_all)
        self.deselect_all.clicked.connect(self._deselect_all)

        self._annotation_color = QColor("red")
        self._create_zoom_toolbar()
        self.drawing_toolbar = self._create_drawing_toolbar()

        title = repr(self.parent())
        self.title_line_edit.setPlaceholderText(title)

        self.combo_title_style.currentTextChanged.connect(self._show_title)
        style = get_config()["logbook_header_style"] or "No title"
        self.combo_title_style.setCurrentText(style)

        self.add_stream.clicked.connect(self.add_destination)
        self.remove_stream.clicked.connect(self.remove_destination)
        self.add_stream.setIcon(icons.plus)
        self.remove_stream.setIcon(icons.minus)

        self.add_destination()

    # -------------------------------------------------------------
    # Karabo Events

    def _event_destinations(self, data):
        """Show the available logbooks for the instrument, in a combobox.

        :param data: A HashList with available LogBooks: Hash with the
            following fields:
                "name" : Logbook identifier
                "destination": Url or folder of the logbook.
        """
        self.data = data

        previous_stream = get_config()["logbook_stream"]
        previous_topic = get_config()["logbook_topic"] or "Logbook"

        for widget in self._destinations:
            widget.update_destinations(data)
            widget.combo_stream.setCurrentText(previous_stream)
            widget.combo_topic.setCurrentText(previous_topic)

        self._update_save_button()

    def _event_network(self, data):
        if not data.get("status"):
            self.close()

    # -----------------------------------------------------------------------
    # Qt Slots

    @Slot()
    def zoom_in(self):
        value = float(self.zoom_factor_edit.text()) * ZOOM_INCREMENT
        self._apply_zoom(value)

    @Slot()
    def zoom_out(self):
        value = float(self.zoom_factor_edit.text()) * (1 / ZOOM_INCREMENT)
        self._apply_zoom(value)

    @Slot()
    def change_zoom_factor(self):
        value = self.zoom_factor_edit.text()
        self.zoom_factor_edit.text()

        factor = float(value) / 100
        self._scale_image(factor=factor)

    def _apply_zoom(self, value):
        if not MIN_ZOOM_FACTOR < value < MAX_ZOOM_FACTOR:
            return
        factor = value / 100
        self._scale_image(factor)
        value = round(value, 1)
        self.zoom_factor_edit.setText(str(value))

    @Slot()
    def _select_all(self):
        for checkbox in self.checkboxes:
            checkbox.setChecked(True)
        self._update_save_button()

    @Slot()
    def _deselect_all(self):
        for checkbox in self.checkboxes:
            checkbox.setChecked(False)
        self._update_save_button()

    @Slot(int)
    def _on_datatype_changed(self, index):
        self.stackedWidget.setCurrentIndex(index)
        self._update_save_button()

    @Slot()
    def _update_save_button(self):
        """Enable the Save button. The button is disabled if

        - no topic is provided
        - dataType is hash and not at least one property is selected
        """

        if not self._destinations:
            self.ok_button.setEnabled(False)
            return
        topic = self._destinations[0].topic
        if not topic:
            self.ok_button.setEnabled(False)
            return

        enabled = self.stackedWidget.currentWidget() == self.image_page
        if not enabled:
            enabled = any(checkbox.isChecked() for checkbox in self.checkboxes)
        self.ok_button.setEnabled(enabled)

    @Slot(bool)
    def _set_drawing_mode(self, enabled):
        self.drawing_toolbar.setVisible(enabled)
        if not enabled:
            self.canvas.set_drawing_tool(None)

    @Slot(str)
    def _show_title(self, text):
        """Do not show the title widget when 'No title' option is selected """
        index = 0 if text == "No title" else 1
        self.title_stackedWidget.setCurrentIndex(index)

    @Slot()
    def add_destination(self):
        """Add a new logbook destination"""
        widget = self._create_destination_widget()
        if self.data is not None:
            widget.update_destinations(self.data)

        # Select the same stream/topic as in the default destination widget.
        destination = self._destinations[0]
        widget.select_destination(destination.stream, destination.topic)

        self.stream_widget.adjustSize()
        self._enable_destination_buttons()

    @Slot()
    def remove_destination(self):
        """Remove the last logbook destination"""
        widget = self._destinations[-1]
        self.destinations_layout.removeWidget(widget)
        widget.deleteLater()
        self.stream_widget.adjustSize()
        self._destinations.remove(widget)
        self._enable_destination_buttons()

    # -----------------------------------------------------------------------
    # Internal Interface

    def _load_table(self):
        data = self.parent().info()
        if data is None:
            return

        self.table.setColumnCount(2)
        self.table.setHorizontalHeaderLabels(["Select", "Property"])
        self.table.setSelectionBehavior(self.table.SelectRows)

        row = 0
        for (k, v, _) in Hash.flat_iterall(data, empty=False):
            if isinstance(v, (bytes, bytearray)):
                continue
            self.table.insertRow(row)

            checkbox = QCheckBox()
            checkbox.setChecked(True)
            checkbox.setStyleSheet(CHECKBOX_STYLE)
            checkbox.toggled.connect(self._update_save_button)
            self.checkboxes.append(checkbox)
            self.table.setCellWidget(row, 0, checkbox)

            item = QTableWidgetItem(k)
            flags = item.flags()
            flags &= ~ Qt.ItemIsEditable
            item.setFlags(flags)
            item.setData(Qt.UserRole, v)
            self.table.setItem(row, 1, item)

            row += 1

    def _scale_image(self, factor):
        self.view.resetTransform()
        self.view.scale(factor, factor)

    def _extract_panel_image(self):
        image_byte = QByteArray()
        io_buffer = QBuffer(image_byte)
        io_buffer.open(QIODevice.WriteOnly)
        image_format = "png"
        pixmap = self.canvas.pixmap_item
        if pixmap is None:
            pixmap = QPixmap(self.pixmap.size())
        # To keep the original size of the image
        pixmap.setDevicePixelRatio(1)
        with QPainter(pixmap) as painter:
            self.canvas.render(painter, source=self.canvas.sceneRect(),
                               mode=Qt.KeepAspectRatio)
        pixmap.save(io_buffer, image_format)
        return {"data": create_base64image(image_format, image_byte),
                "dataType": "text_image",
                "caption": self.caption_edit.toPlainText()}

    def _extract_panel_data(self):
        h = Hash()
        for row, checkbox in enumerate(self.checkboxes):
            if checkbox.isChecked():
                item = self.table.item(row, 1)
                name = item.data(Qt.DisplayRole)
                value = item.data(Qt.UserRole)
                h[name] = value
        return {"data": h, "dataType": "hash"}

    def _create_logbook_data(self):
        panel = self.parent()
        stack = self.combo_datatype.currentText()
        info = {}
        header_type = "none"

        title = self.title_line_edit.text().strip()
        if not title:
            title = repr(panel)

        if self.combo_title_style.currentText() != "No title":
            header_type = self.combo_title_style.currentText().lower()

        if stack == LOGBOOK_DATA:
            info["title"] = f"Data: {title}"
            info.update(self._extract_panel_data())
            info["uploadFile"] = self.upload_file.isChecked()
        elif stack == LOGBOOK_IMAGE:
            info["title"] = f"Image: {title}"
            info.update(self._extract_panel_image())

        info["headerType"] = header_type

        return info

    @Slot()
    def _fit_in_view(self):
        """Adjust the image's size to fit the QGraphicsView, ensuring the
        entire image is visible."""
        scene_rect = self.view.sceneRect()
        self.view.fitInView(scene_rect, Qt.KeepAspectRatio)

        # Show the current zoom factor of the image.
        transform = self.view.transform()
        width_scale = transform.m11()
        height_scale = transform.m22()
        zooming_factor = ((width_scale + height_scale) / 2) * 100
        self.zoom_factor_edit.setText(str(round(zooming_factor, 1)))

    def _create_destination_widget(self):
        destination_widget = DestinationWidget(parent=self.stream_widget)
        self.destinations_layout.addWidget(destination_widget)
        self._destinations.append(destination_widget)
        return destination_widget

    def _enable_destination_buttons(self):
        """Enable/Disable the Add/Remove Destination buttons- to allow min 1
        and max 3 destinations. """
        size = len(self._destinations)
        add_enabled = size < MAX_ALLOWED_LOGBOOK
        remove_enabled = size > 1
        self.add_stream.setEnabled(add_enabled)
        self.remove_stream.setEnabled(remove_enabled)

    # -----------------------------------------------------------------------
    # Public interface

    def done(self, result):
        if result:
            self.request_save()
            self._save_configs()
        unregister_from_broadcasts(self._event_map)
        return super().done(result)

    def request_save(self):
        """Request saving a Hash to the KaraboLogBook"""

        info = self._create_logbook_data()
        info["destinations"] = self._get_destinations()
        get_network().onSaveLogBook(**info)

    def _get_destinations(self):
        """Create a HashList from the selected Logbook stream and topic
        names in the GUI"""
        destinations = HashList()
        for widget in self._destinations:
            h = Hash("stream", widget.stream, "topic", widget.topic)
            destinations.append(h)
        return destinations

    def _save_configs(self):
        """Save the singleton configurations."""
        widget = self._destinations[0]
        get_config()["logbook_stream"] = widget.stream
        get_config()["logbook_topic"] = widget.topic
        style = self.combo_title_style.currentText()
        get_config()["logbook_header_style"] = style

    def _create_zoom_toolbar(self):
        """ Create Toolbar with zooming options"""

        zoom_toolbar = ToolBar(parent=self)
        zoom_in_action = QAction(icons.zoomIn, "Zoom In", self)
        zoom_in_action.triggered.connect(self.zoom_in)

        zoom_out_action = QAction(icons.zoomOut, "Zoom Out", self)
        zoom_out_action.triggered.connect(self.zoom_out)
        zoom_toolbar.addAction(zoom_in_action)
        zoom_toolbar.addAction(zoom_out_action)
        size = QSize(23, 23)
        zoom_toolbar.setIconSize(size)
        zoom_toolbar.add_expander()
        self.toolbar_layout.insertWidget(0, zoom_toolbar)

    def _create_drawing_toolbar(self):
        """Create a toolbar with annotate tools """
        toolbar = ToolBar(parent=self)
        self.action_group_draw = QActionGroup(self)
        text = "Annotations cannot be \nremoved after cropping"
        warning_text = QLabel(text, parent=toolbar)
        font = warning_text.font()
        font.setPointSize(10)
        font.setItalic(True)
        warning_text.setFont(font)
        self._warning_action = toolbar.addWidget(warning_text)
        self._warning_action.setVisible(False)

        color_action = QAction(parent=self)
        color_action.setToolTip("Change annotation color")
        color_action.triggered.connect(self._change_color)
        toolbar.addAction(color_action)
        self.action_group_draw.addAction(color_action)
        self._color_action = color_action
        self._set_button_color()

        for tool_factory in get_tools():
            action = QAction(parent=self)
            action.setIcon(tool_factory.icon)
            action.setToolTip(tool_factory.tooltip)
            action.setCheckable(True)
            tool = tool_factory.drawing_tool()
            action.triggered.connect(partial(self.activate_tool, tool))
            toolbar.addAction(action)
            self.action_group_draw.addAction(action)

        self.action_group_draw.setExclusive(True)
        self.toolbar_layout.addWidget(toolbar)
        return toolbar

    @Slot()
    def _change_color(self):
        """Show a color dialog to choose the annotation color """
        self._warning_action.setVisible(False)
        color = QColorDialog.getColor(
            initial=self._annotation_color, parent=self)
        # To avoid the dialog hiding behind the main window
        if IS_MAC_SYSTEM:
            self.raise_()
        if color.isValid():
            self._annotation_color = color
            self._set_button_color()
            self.canvas.set_drawing_color(color)

    def _set_button_color(self):
        """Set the color on the color button."""
        pixmap = QPixmap(16, 16)
        pixmap.fill(self._annotation_color)
        self._color_action.setIcon(QIcon(pixmap))

    @Slot()
    def activate_tool(self, tool):
        self.view.setCursor(Qt.CrossCursor)
        self.canvas.set_drawing_tool(tool)
        self.canvas.set_drawing_color(self._annotation_color)
        warning_visible = isinstance(tool, CropTool)
        self._warning_action.setVisible(warning_visible)

    @Slot()
    def reset_tool(self):
        for action in self.action_group_draw.actions():
            action.setChecked(False)
        self.view.setCursor(Qt.ArrowCursor)
        self._warning_action.setVisible(False)



import numpy as np

from karabo.native import (
    Assignment, Bool, Configurable, Hash, Int8, Node, Slot, String)
from karabogui.binding.api import (
    DeviceProxy, apply_default_configuration, build_binding)
from karabogui.dialogs.init_configuration_preview import (
    InitConfigurationPreviewDialog, extract_defaults)
from karabogui.testing import singletons
from karabogui.topology.api import SystemTopology


class ANode(Configurable):
    name = String(defaultValue="A Node")
    integer = Int8(defaultValue=1)
    flag = Bool(defaultValue=True)


class Object(Configurable):
    val = String(defaultValue="FooString")
    intValue = Int8(defaultValue=10)
    intNoDefault = Int8()
    aNode = Node(ANode, displayedName="TestNode")
    internal = Int8(defaultValue=1, assignment=Assignment.INTERNAL)

    @Slot()
    async def dummySlot(self):
        pass


def get_default_values():
    node = Hash("name", "A Node", "integer", np.int8(1), "flag", True)
    return Hash("val", "FooString", "intValue", np.int8(10),
                "aNode", node)


init_config = """
val
New text
aNode

integer
20
"""

init_config_with_defaults = """
val
New text
intValue
10
aNode

name
A Node
integer
20
flag
True
"""


def test_extract_default_value_configuration():
    schema = Object.getClassSchema()
    binding = build_binding(schema)
    apply_default_configuration(binding)
    expected = get_default_values()
    default_values = extract_defaults(binding)
    assert default_values == expected


def test_init_configuration_preview_dialog(gui_app):
    """Test the InitConfigurationPreviewDialog with a sample configuration."""
    schema = Object.getClassSchema()
    binding = build_binding(schema)
    apply_default_configuration(binding)
    proxy = DeviceProxy(binding=binding)
    proxy.server_id = "swerver"

    node = Hash("integer", np.int8(20))
    config = Hash("val", "New text", "aNode", node)

    topology = SystemTopology()
    topology._class_proxies = {
        (proxy.server_id, proxy.binding.class_id): proxy}
    with singletons(topology=topology):
        dialog = InitConfigurationPreviewDialog(proxy, config,
                                                title="Test Dialog")
        assert dialog.ui_configuration_text.toPlainText() == init_config
        assert dialog.ui_full_configuration_text.toPlainText() == (
            init_config_with_defaults)

        # Verify the configuration to apply.
        assert dialog.configuration_to_apply == config
        dialog.ui_configuration_selector.setCurrentIndex(1)
        full_config = extract_defaults(binding)
        full_config.merge(config)
        assert dialog.configuration_to_apply == full_config



#############################################################################
# Author: <chen.xu@xfel.eu> & <dennis.goeries@xfel.eu>
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from collections import deque
from functools import partial
from itertools import cycle, product
from weakref import WeakValueDictionary

import numpy as np
from qtpy.QtWidgets import QWidget
from traits.api import Bool, Callable, Dict, Instance, List

import karabogui.icons as icons
from karabo.common.scenemodel.api import (
    MultiCurveGraphModel, build_model_config)
from karabogui.binding.api import (
    BoolBinding, FloatBinding, IntBinding, PropertyProxy)
from karabogui.controllers.api import (
    BaseBindingController, register_binding_controller)
from karabogui.graph.common.api import create_button, get_pen_cycler
from karabogui.graph.plots.api import KaraboPlotView

NUMERICAL_BINDINGS = (BoolBinding, FloatBinding, IntBinding)
MAXNUMPOINTS = 1000


def is_compatible(binding):
    """Reject a boolean for the x-axis"""
    return not isinstance(binding, BoolBinding)


@register_binding_controller(ui_name='Multi-Curve Graph',
                             klassname='MultiCurveGraph',
                             binding_type=NUMERICAL_BINDINGS,
                             is_compatible=is_compatible,
                             can_show_nothing=False)
class DisplayMultiCurveGraph(BaseBindingController):
    # The scene model class used by this controller
    model = Instance(MultiCurveGraphModel, args=())
    # Internal traits
    _mpl_widget = Instance(QWidget)
    _reset_proxy = Instance(PropertyProxy)
    _y_proxies = List(Instance(PropertyProxy))
    _x_values = Instance(deque, kw={'maxlen': MAXNUMPOINTS})
    _y_values = Dict()  # {box.key(): deque of values}
    _last_values = Dict()  # used to synchronize x and y values
    _draw_start = Bool(False)  # indicate there is at least one x-y pair
    _resetbox_linked = Callable

    # The associated plot curves
    _curves = Instance(WeakValueDictionary, args=())
    _pens = Instance(cycle, allow_none=False)

    def __pens_default(self):
        return get_pen_cycler()

    def create_widget(self, parent):
        widget = KaraboPlotView(parent=parent)
        widget.add_legend(visible=False)
        widget.add_cross_target()
        widget.enable_data_toggle(True)
        toolbar = widget.add_toolbar()

        _btn_reset = create_button(checkable=False,
                                   icon=icons.reset,
                                   tooltip="Reset the plot",
                                   on_clicked=partial(self._reset_plot, True))
        toolbar.add_button(_btn_reset)
        widget.stateChanged.connect(self._change_model)

        # A closure to change the button style once a boolean is linked
        def _changestyle(default=False):
            icon = icons.reset if default else icons.resetTilted
            _btn_reset.setIcon(icon)

        self._resetbox_linked = _changestyle

        self._last_values[self.proxy] = None
        self._draw_start = True
        widget.restore(build_model_config(self.model))
        return widget

    def binding_update(self, proxy):
        self.add_proxy(proxy)

    def add_proxy(self, proxy):
        binding = proxy.binding
        if binding is None:
            # Wait for a new cycle
            return True

        if self._reset_proxy is None and isinstance(binding, BoolBinding):
            # if the resetbox is set, new boolean box goes to yvalues
            self._reset_proxy = proxy
            self._resetbox_linked()
            return True

        if proxy in self._last_values:
            return False

        self._last_values[proxy] = None
        self._y_proxies.append(proxy)
        # request a new curve for each new ybox, use proxy.key for the
        # curve name as this is unique
        name = proxy.key
        self._curves[proxy] = self.widget.add_curve_item(
            name=name, pen=next(self._pens))
        self._y_values[proxy] = deque(maxlen=MAXNUMPOINTS)

        # We can show the legend already here!
        self.widget.set_legend(True)

        return True

    def remove_proxy(self, proxy):
        if proxy is self._reset_proxy:
            self._reset_proxy = None
            self._resetbox_linked(default=True)
            return True
        if proxy in self._y_proxies:
            self._y_proxies.remove(proxy)

            item = self._curves.pop(proxy)
            self.widget.remove_item(item)
            self._y_values.pop(proxy)
            self._last_values.pop(proxy)
            if not self._curves:
                self.widget.set_legend(False)
            return True
        return False

    def value_update(self, proxy):
        value = proxy.value
        if proxy is self._reset_proxy:
            self._reset_plot(reset=value)
            return

        synchronizer = self._last_values
        synchronizer[proxy] = value
        if self._draw_start and None not in synchronizer.values():
            self._update_curve()

    # -------------------------------------------------------------------
    # private functions

    def _reset_plot(self, reset=True):
        """clear the plot"""
        if not reset:
            # when adding a boolean as resetbox, it may send a False value
            return

        self._last_values = {k: None for k in self._last_values.keys()}
        self._x_values.clear()
        for proxy, arr in self._y_values.items():
            arr.clear()
            self._curves[proxy].setData(self._x_values, arr)

    def _update_curve(self):
        """Update all curves, clear lastvalues buffer (set all values to None).
        This is called when xbox and all y boxes produce a not None value.
        """
        lastvalues = self._last_values
        for proxy, value in lastvalues.items():
            if proxy is self.proxy:
                self._x_values.append(value)
            else:
                self._y_values[proxy].append(value)

        # product pairs xvalues to each ('proxy', yvalues)
        for xdata, (proxy, ydata) in product((self._x_values,),
                                             self._y_values.items()):
            # later added y may have shorter length
            npx, npy = _convert_data(xdata, ydata)
            self._curves[proxy].setData(npx, npy)
        self._last_values = {k: None for k in lastvalues.keys()}

    # ----------------------------------------------------------------
    # Qt Slots

    def _change_model(self, content):
        self.model.trait_set(**content)


def _convert_data(xdata, ydata):
    """Convert x y deque array to numpy array for plotting.
    """
    # There might be np.nan in the array
    npx = np.array(xdata, dtype=np.float64)
    npy = np.array(ydata, dtype=np.float64)
    if npx.size != npy.size:
        # trim npx if necessary, len(npx) >= len(npy) always True
        npx = npx[-npy.size:]
    # get a mask to get rid of np.nan in y
    mask = np.isfinite(npy)
    return npx[mask], npy[mask]



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

from traits.api import Instance, WeakRef

from karabo.common.scenemodel.api import (
    VectorFillGraphModel, build_model_config)
from karabogui.binding.api import VectorNumberBinding
from karabogui.controllers.api import (
    BaseBindingController, register_binding_controller)
from karabogui.graph.plots.api import KaraboPlotView, generate_down_sample


@register_binding_controller(ui_name='Vector Fill Graph',
                             klassname='VectorFillGraph',
                             binding_type=VectorNumberBinding,
                             can_show_nothing=False)
class DisplayVectorFillGraph(BaseBindingController):
    """The VectorFill Graph controller for display of vector data

    This widget fills the background between curve and axis.
    """
    model = Instance(VectorFillGraphModel, args=())
    _plot = WeakRef(object)

    def create_widget(self, parent):
        widget = KaraboPlotView(parent=parent)
        widget.add_cross_target()
        widget.add_toolbar()
        widget.stateChanged.connect(self._change_model)
        self._plot = widget.add_curve_fill()
        widget.restore(build_model_config(self.model))

        return widget

    # ----------------------------------------------------------------

    def value_update(self, proxy):
        value = proxy.value
        # This plotItem is special!
        _, y = generate_down_sample(value, threshold=2000, rect=None)
        self._plot.setData(y)

    # ----------------------------------------------------------------
    # Qt Slots

    def _change_model(self, content):
        self.model.trait_set(**content)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QAction, QDialog, QFrame
from traits.api import Instance

from karabo.common.scenemodel.api import DisplayListModel
from karabogui.binding.api import (
    VectorBinding, VectorCharBinding, VectorHashBinding, get_binding_value)
from karabogui.controllers.api import (
    BaseBindingController, register_binding_controller)
from karabogui.dialogs.format_label import FormatLabelDialog
from karabogui.fonts import get_font_size_from_dpi
from karabogui.indicators import ALL_OK_COLOR
from karabogui.util import generateObjectName
from karabogui.widgets.hints import ElidingLabel


def _is_compatible(binding):
    return not isinstance(binding, (VectorCharBinding, VectorHashBinding))


@register_binding_controller(ui_name="List", is_compatible=_is_compatible,
                             klassname="DisplayList", priority=10,
                             binding_type=VectorBinding)
class DisplayList(BaseBindingController):
    model = Instance(DisplayListModel, args=())

    def create_widget(self, parent):
        widget = ElidingLabel(parent)
        widget.setFrameStyle(QFrame.Box | QFrame.Plain)
        widget.setFocusPolicy(Qt.NoFocus)
        objectName = generateObjectName(self)
        widget.setObjectName(objectName)
        style_sheet = (f"QWidget#{objectName}" +
                       f" {{ background-color : rgba{ALL_OK_COLOR}; }}")
        widget.setStyleSheet(style_sheet)
        action = QAction("Format field...", widget)
        action.triggered.connect(self._format_field)
        widget.addAction(action)

        return widget

    def value_update(self, proxy):
        value = get_binding_value(proxy, [])
        self.widget.setText(",".join(str(v) for v in value))

    def _format_field(self):
        dialog = FormatLabelDialog(font_size=self.model.font_size,
                                   font_weight=self.model.font_weight,
                                   parent=self.widget)
        if dialog.exec() == QDialog.Accepted:
            self.model.trait_set(
                font_size=dialog.font_size, font_weight=dialog.font_weight)
            self._apply_format()

    def _apply_format(self, widget=None):
        if widget is None:
            widget = self.widget

        # Apply font formatting
        font = widget.font()
        font.setPointSize(get_font_size_from_dpi(self.model.font_size))
        font.setBold(self.model.font_weight == "bold")
        widget.setFont(font)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from traits.api import Constant, Dict, HasStrictTraits, Property, String


class BaseStats(HasStrictTraits):

    # Statistics from the analyzer
    stats = Dict

    # HTML table body
    html_format = Constant("<table style='font-size: 9px'>"
                           "{contents}</table>")

    # HTML table contents
    html_table = Constant('')

    # Rendered HTML
    html = Property(String)

    def _get_html(self):
        """Subclass to implement specific html composition"""
        html = ''
        if self.stats:
            html = self.html_table.format(**self.stats)
        return html

    # -----------------------------------------------------------------------
    # Private methods

    @staticmethod
    def _to_string(value):
        """Converts the numerical value into an readable number"""
        if value is None:
            return "-"
        return f"{value:.2f}".rstrip('0').rstrip('.')



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from traits.api import pop_exception_handler, push_exception_handler

from karabo.common.api import InstanceStatus
from karabogui.itemtypes import NavigationItemTypes
from karabogui.testing import system_hash, system_hash_server_and_plugins

from ..system_tree import SystemTree, SystemTreeNode


def setUp():
    push_exception_handler(lambda *args: None, reraise_exceptions=True)


def tearDown():
    pop_exception_handler()


def test_tree_node_basics():
    empty = SystemTreeNode(
        node_id='node_id', path='path',
        status=InstanceStatus.NONE,
        children=[]
    )

    assert empty.child('a_child') is None
    assert empty.info() is None
    assert empty.row() == 0


def test_tree_node_levels():
    root = SystemTreeNode()
    level_0 = SystemTreeNode(node_id='level_0', parent=root, level=0)
    root.children.append(level_0)
    level_1 = SystemTreeNode(node_id='level_1', parent=level_0, level=1)
    level_0.children.append(level_1)
    level_2 = SystemTreeNode(node_id='level_2', parent=level_1, level=2)
    level_1.children.append(level_2)
    level_3 = SystemTreeNode(node_id='level_3', parent=level_2, level=3)
    level_2.children.append(level_3)

    assert root.child('level_0') is level_0
    assert level_0.child('level_1') is level_1 and level_0.row() == 0
    assert level_1.child('level_2') is level_2 and level_1.row() == 0
    assert level_2.child('level_3') is level_3 and level_2.row() == 0

    level_0_info = level_0.info()
    assert level_0_info['type'] == NavigationItemTypes.HOST
    level_1_info = level_1.info()
    assert level_1_info['type'] == NavigationItemTypes.SERVER
    assert level_1_info['serverId'] == 'level_1'
    assert level_1_info['log'] == 'INFO'
    level_2_info = level_2.info()
    assert level_2_info['type'] == NavigationItemTypes.CLASS
    assert level_2_info['serverId'] == 'level_1'
    assert level_2_info['classId'] == 'level_2'
    assert level_2_info['deviceId'] == ''
    level_3_info = level_3.info()
    assert level_3_info['type'] == NavigationItemTypes.DEVICE
    assert level_3_info['classId'] == 'level_2'
    assert level_3_info['deviceId'] == 'level_3'


def test_tree_basics():
    tree = SystemTree()
    assert len(tree.find('anything')) == 0

    sys_hash = system_hash()
    names = ('swerver', 'divvy', 'macdonald')

    for node_id in names:
        assert len(tree.find(node_id)) == 0

    tree.initialize(sys_hash)
    for node_id in names:
        assert len(tree.find(node_id)) > 0, node_id

    tree.clear_all()
    for node_id in names:
        assert len(tree.find(node_id)) == 0


def test_tree_find():
    tree = SystemTree()
    names = ('FooClass', 'BarClass')

    for node_id in names:
        assert len(tree.find(node_id)) == 0

    h = system_hash_server_and_plugins()
    tree.update(h)
    assert len(tree.find('BarClass')) == 1
    assert len(tree.find('barclass', case_sensitive=False)) == 1
    assert len(tree.find('Bar')) == 1
    assert len(tree.find('FooClass')) == 1
    assert len(tree.find('BarClass')) == 1
    assert len(tree.find('BlahClass')) == 0
    kwargs = {'case_sensitive': True,
              'use_reg_ex': True}
    assert len(tree.find('(.*)Class', **kwargs)) == 2
    assert len(tree.find('(.*)class', **kwargs)) == 1
    kwargs['case_sensitive'] = False
    kwargs['use_reg_ex'] = True
    assert len(tree.find('(.*)fooclass', **kwargs)) == 1
    assert len(tree.find('HooClass', **kwargs)) == 0
    kwargs['full_match'] = True
    assert len(tree.find('HooClass', **kwargs)) == 0
    assert len(tree.find('HooClass_0', **kwargs)) == 0
    assert len(tree.find('HooClass_1', **kwargs)) == 0


def test_tree_clear_existing():
    tree = SystemTree()
    names = ('swerver', 'divvy', 'macdonald', 'orphan')
    sys_hash = system_hash()

    tree.update(sys_hash)
    tree.clear_existing(sys_hash)
    for node_id in names:
        assert len(tree.find(node_id)) == 0



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on February 1, 2012
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtWidgets import QAction, QVBoxLayout, QWidget

from karabogui import icons
from karabogui.events import (
    KaraboEvent, register_for_broadcasts, unregister_from_broadcasts)
from karabogui.widgets.log import LogWidget
from karabogui.widgets.toolbar import ToolBar

from .base import BasePanelWidget


class LoggingPanel(BasePanelWidget):
    def __init__(self):
        super().__init__("Log", allow_closing=True)
        # Register for broadcast events.
        self.event_map = {
            KaraboEvent.LogMessages: self._event_log_messages,
            KaraboEvent.NetworkConnectStatus: self._event_network,
        }
        register_for_broadcasts(self.event_map)

    def get_content_widget(self):
        """Returns a QWidget containing the main content of the panel.
        """
        widget = QWidget(self)

        self._log_widget = LogWidget(parent=widget)
        mainLayout = QVBoxLayout(widget)
        mainLayout.setContentsMargins(5, 5, 5, 5)
        mainLayout.addWidget(self._log_widget)

        return widget

    def toolbars(self):
        """This should create and return one or more `ToolBar` instances needed
        by this panel.
        """
        text = "Save log data to file"
        self.__acSaveLog = QAction(icons.save, "&Save log data (.log)", self)
        self.__acSaveLog.setToolTip(text)
        self.__acSaveLog.setStatusTip(text)
        self.__acSaveLog.triggered.connect(self._log_widget.onSaveToFile)

        text = "Clear log"
        self.__acClearLog = QAction(icons.editClear, text, self)
        self.__acClearLog.setToolTip(text)
        self.__acClearLog.setStatusTip(text)
        self.__acClearLog.triggered.connect(self._log_widget.onClearLog)

        toolBar = ToolBar(parent=self)
        toolBar.addAction(self.__acSaveLog)
        toolBar.addAction(self.__acClearLog)
        return [toolBar]

    def closeEvent(self, event):
        """Unregister from broadcast events, tell main window to enable
        the button to add me back.
        """
        super().closeEvent(event)
        if event.isAccepted():
            unregister_from_broadcasts(self.event_map)
            self.signalPanelClosed.emit(self.windowTitle())

    # -----------------------------------------------------------------------
    # Karabo Events

    def _event_log_messages(self, data):
        messages = data.get('messages', [])
        self._log_widget.onLogDataAvailable(messages)

    def _event_network(self, data):
        if not data['status']:
            # on False status we only clear the logs
            self._log_widget.onClearLog()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import os.path as op

from qtpy import uic
from qtpy.QtCore import Qt
from qtpy.QtGui import QPixmap
from qtpy.QtWidgets import QDialog

from karabogui import const, icons
from karabogui.singletons.api import get_manager, get_network

from .utils import get_dialog_ui


class _PatternMatcher:
    """A tiny state machine which watches for a single pattern.

    Useful for watching for specific key sequences...
    """

    def __init__(self, pattern):
        self.pattern = pattern
        self.index = 0

    def check(self, letter):
        if letter == self.pattern[self.index]:
            self.index += 1
            if self.index == len(self.pattern):
                self.index = 0
                return True
        else:
            self.index = 0

        return False


class AboutDialog(QDialog):
    """The about box for our application.

    NOTE: We watch for "cheat codes" here to enable/disable certain application
    features.
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        uic.loadUi(get_dialog_ui('about.ui'), self)
        self.setAttribute(Qt.WA_DeleteOnClose)

        image_path = op.join(op.dirname(icons.__file__), 'tunnel.png')
        tunnel_img = QPixmap(image_path)
        self.imgLabel.setPixmap(tunnel_img)

        txt = self.aboutText.text()
        self.aboutText.setText(
            txt.replace('$KARABO_VERSION', const.GUI_VERSION_DETAILED, 1))

        # Pattern matchers for a specific key combos
        self._cheat_codes = {
            _PatternMatcher('chooch'):
                lambda: get_network().togglePerformanceMonitor(),
            _PatternMatcher('bigdata'):
                lambda: get_manager().toggleBigDataPerformanceMonitor()
        }

    def keyPressEvent(self, event):
        text = event.text()
        for matcher, trigger in self._cheat_codes.items():
            if matcher.check(text):
                trigger()
        return super().keyPressEvent(event)
